@@ -199,14 +199,16 @@ datadir_iter_new(const char *path, bool skip_first_level = true)
 	datadir_iter_t *it;
 
 	it = static_cast<datadir_iter_t *>(malloc(sizeof(datadir_iter_t)));
+	if (!it)
+		goto error;
 	memset(it, 0, sizeof(datadir_iter_t));
 
 	pthread_mutex_init(&it->mutex, NULL);
 	it->datadir_path = strdup(path);
 
-	it->dir = os_file_opendir(it->datadir_path, TRUE);
+	it->dir = os_file_opendir(it->datadir_path);
 
-	if (it->dir == NULL) {
+	if (it->dir == IF_WIN(INVALID_HANDLE_VALUE, nullptr)) {
 
 		goto error;
 	}
@@ -237,11 +239,9 @@ bool
 datadir_iter_next_database(datadir_iter_t *it)
 {
 	if (it->dbdir != NULL) {
-		if (os_file_closedir(it->dbdir) != 0) {
-
+		if (os_file_closedir_failed(it->dbdir)) {
 			msg("Warning: could not"
 			      " close database directory %s", it->dbpath);
-
 			it->err = DB_ERROR;
 
 		}
@@ -286,10 +286,9 @@ datadir_iter_next_database(datadir_iter_t *it)
 
 		/* We want wrong directory permissions to be a fatal error for
 		XtraBackup. */
-		it->dbdir = os_file_opendir(it->dbpath, TRUE);
-
-		if (it->dbdir != NULL) {
+		it->dbdir = os_file_opendir(it->dbpath);
 
+		if (it->dir != IF_WIN(INVALID_HANDLE_VALUE, nullptr)) {
 			it->is_file = false;
 			return(true);
 		}
@@ -729,9 +728,9 @@ directory_exists(const char *dir, bool create)
 	}
 
 	/* could be symlink */
-	os_dir = os_file_opendir(dir, FALSE);
+	os_dir = os_file_opendir(dir);
 
-	if (os_dir == NULL) {
+	if (os_dir == IF_WIN(INVALID_HANDLE_VALUE, nullptr)) {
 		my_strerror(errbuf, sizeof(errbuf), my_errno);
 		msg("Can not open directory %s: %s", dir,
 			errbuf);
@@ -759,9 +758,9 @@ directory_exists_and_empty(const char *dir, const char *comment)
 		return(false);
 	}
 
-	os_dir = os_file_opendir(dir, FALSE);
+	os_dir = os_file_opendir(dir);
 
-	if (os_dir == NULL) {
+	if (os_dir == IF_WIN(INVALID_HANDLE_VALUE, nullptr)) {
 		msg("%s can not open directory %s", comment, dir);
 		return(false);
 	}
@@ -1819,26 +1818,11 @@ copy_back()
 	dst_dir = (srv_log_group_home_dir && *srv_log_group_home_dir)
 		? srv_log_group_home_dir : mysql_data_home;
 
-	/* --backup generates a single LOG_FILE_NAME, which we must copy
-	if it exists. */
+	/* --backup generates a single ib_logfile0, which we must copy. */
 
 	ds_data = ds_create(dst_dir, DS_TYPE_LOCAL);
-	MY_STAT stat_arg;
-	if (!my_stat(LOG_FILE_NAME, &stat_arg, MYF(0)) || !stat_arg.st_size) {
-		/* After completed --prepare, redo log files are redundant.
-		We must delete any redo logs at the destination, so that
-		the database will not jump to a different log sequence number
-		(LSN). */
-
-		char filename[FN_REFLEN];
-		snprintf(filename, sizeof filename, "%s/%s0", dst_dir,
-			 LOG_FILE_NAME_PREFIX);
-		unlink(filename);
-		snprintf(filename, sizeof filename, "%s/%s101", dst_dir,
-			 LOG_FILE_NAME_PREFIX);
-		unlink(filename);
-	} else if (!(ret = copy_or_move_file(LOG_FILE_NAME, LOG_FILE_NAME,
-					     dst_dir, 1))) {
+	if (!(ret = copy_or_move_file(LOG_FILE_NAME, LOG_FILE_NAME,
+				      dst_dir, 1))) {
 		goto cleanup;
 	}
 	ds_destroy(ds_data);
@@ -2111,7 +2095,9 @@ decrypt_decompress()
 */
 static bool backup_files_from_datadir(const char *dir_path)
 {
-	os_file_dir_t dir = os_file_opendir(dir_path, TRUE);
+	os_file_dir_t dir = os_file_opendir(dir_path);
+	if (dir == IF_WIN(INVALID_HANDLE_VALUE, nullptr)) return false;
+
 	os_file_stat_t info;
 	bool ret = true;
 	while (os_file_readdir_next_file(dir_path, dir, &info) == 0) {
