@@ -19,6 +19,9 @@
 #include "sql_table.h"                       // mysql_alter_table,
                                              // mysql_exchange_partition
 #include "sql_alter.h"
+#include "rpl_mi.h"
+#include "slave.h"
+#include "debug_sync.h"
 #include "wsrep_mysqld.h"
 
 Alter_info::Alter_info(const Alter_info &rhs, MEM_ROOT *mem_root)
@@ -254,16 +257,8 @@ Alter_info::algorithm(const THD *thd) const
 
 
 Alter_table_ctx::Alter_table_ctx()
-  : implicit_default_value_error_field(NULL),
-    error_if_not_empty(false),
-    tables_opened(0),
-    db(null_clex_str), table_name(null_clex_str), alias(null_clex_str),
-    new_db(null_clex_str), new_name(null_clex_str), new_alias(null_clex_str),
-    storage_engine_partitioned(false),
-    tmp_storage_engine_name_partitioned(false),
-    fk_error_if_delete_row(false), fk_error_id(NULL),
-    fk_error_table(NULL),
-    tmp_table(false)
+  : db(null_clex_str), table_name(null_clex_str), alias(null_clex_str),
+    new_db(null_clex_str), new_name(null_clex_str), new_alias(null_clex_str)
 {
 }
 
@@ -276,13 +271,8 @@ Alter_table_ctx::Alter_table_ctx(THD *thd, TABLE_LIST *table_list,
                                  uint tables_opened_arg,
                                  const LEX_CSTRING *new_db_arg,
                                  const LEX_CSTRING *new_name_arg)
-  : implicit_default_value_error_field(NULL), error_if_not_empty(false),
-    tables_opened(tables_opened_arg),
-    new_db(*new_db_arg), new_name(*new_name_arg),
-    tmp_storage_engine_name_partitioned(false),
-    fk_error_if_delete_row(false), fk_error_id(NULL),
-    fk_error_table(NULL),
-    tmp_table(false)
+  : tables_opened(tables_opened_arg),
+    new_db(*new_db_arg), new_name(*new_name_arg)
 {
   /*
     Assign members db, table_name, new_db and new_name
@@ -517,7 +507,11 @@ bool Sql_cmd_alter_table::execute(THD *thd)
                                    lex->name.str ? lex->name.str
                                    : first_table->table_name.str,
                                    first_table, &alter_info, &keys,
-                                   used_engine ? &create_info : nullptr);
+                                   used_engine ? &create_info : nullptr)
+    {
+      WSREP_WARN("ALTER TABLE isolation failure");
+      DBUG_RETURN(TRUE);
+    }
 
     thd->variables.auto_increment_offset = 1;
     thd->variables.auto_increment_increment = 1;
@@ -559,11 +553,6 @@ bool Sql_cmd_alter_table::execute(THD *thd)
                             lex->ignore, lex->if_exists());
 
   DBUG_RETURN(result);
-#ifdef WITH_WSREP
-wsrep_error_label:
-  WSREP_WARN("ALTER TABLE isolation failure");
-  DBUG_RETURN(TRUE);
-#endif
 }
 
 bool Sql_cmd_discard_import_tablespace::execute(THD *thd)
