@@ -169,8 +169,12 @@ inline void buf_pool_t::delete_from_flush_list_low(buf_page_t *bpage)
 void buf_pool_t::insert_into_flush_list(buf_block_t *block, lsn_t lsn)
 {
   mysql_mutex_assert_not_owner(&mutex);
-  mysql_mutex_assert_owner(&log_sys.flush_order_mutex);
+#ifdef SAFE_MUTEX
+  if (!recv_recovery_is_on())
+    mysql_mutex_assert_owner(&log_sys.flush_order_mutex);
+#endif /* SAFE_MUTEX */
   ut_ad(lsn > 2);
+  static_assert(log_t::FIRST_LSN >= 2, "compatibility");
   ut_ad(!fsp_is_system_temporary(block->page.id().space()));
 
   mysql_mutex_lock(&flush_list_mutex);
@@ -418,7 +422,7 @@ void buf_flush_assign_full_crc32_checksum(byte* page)
 	ut_ad(!corrupted);
 	ut_ad(size == uint(srv_page_size));
 	const ulint payload = srv_page_size - FIL_PAGE_FCRC32_CHECKSUM;
-	mach_write_to_4(page + payload, ut_crc32(page, payload));
+	mach_write_to_4(page + payload, my_crc32c(0, page, payload));
 }
 
 /** Initialize a page for writing to the tablespace.
@@ -573,7 +577,7 @@ static void buf_tmp_reserve_compression_buf(buf_tmp_buffer_t* slot)
   buffer be bigger than input buffer. Adjust the allocated size. */
   ulint size= srv_page_size;
   if (provider_service_lzo->is_loaded)
-    size= LZO1X_1_15_MEM_COMPRESS;
+    size+= LZO1X_1_15_MEM_COMPRESS;
   else if (provider_service_snappy->is_loaded)
     size= snappy_max_compressed_length(size);
   slot->comp_buf= static_cast<byte*>(aligned_malloc(size, srv_page_size));
@@ -599,7 +603,7 @@ static byte* buf_tmp_page_encrypt(ulint offset, const byte* s, byte* d)
     return NULL;
 
   const ulint payload= srv_page_size - FIL_PAGE_FCRC32_CHECKSUM;
-  mach_write_to_4(d + payload, ut_crc32(d, payload));
+  mach_write_to_4(d + payload, my_crc32c(0, d, payload));
 
   srv_stats.pages_encrypted.inc();
   srv_stats.n_temp_blocks_encrypted.inc();
@@ -621,6 +625,7 @@ static byte *buf_page_encrypt(fil_space_t* space, buf_page_t* bpage, byte* s,
   ut_ad(space->id == bpage->id().space());
   ut_ad(!*slot);
 
+  ut_d(fil_page_type_validate(space, s));
   const uint32_t page_no= bpage->id().page_no();
 
   switch (page_no) {
@@ -717,6 +722,7 @@ not_compressed:
 
     /* Workaround for MDEV-15527. */
     memset(tmp + len, 0 , srv_page_size - len);
+    ut_d(fil_page_type_validate(space, tmp));
 
     if (encrypted)
       tmp= fil_space_encrypt(space, page_no, tmp, d);
@@ -724,13 +730,14 @@ not_compressed:
     if (full_crc32)
     {
       static_assert(FIL_PAGE_FCRC32_CHECKSUM == 4, "alignment");
-      mach_write_to_4(tmp + len - 4, ut_crc32(tmp, len - 4));
+      mach_write_to_4(tmp + len - 4, my_crc32c(0, tmp, len - 4));
       ut_ad(!buf_page_is_corrupted(true, tmp, space->flags));
     }
 
     d= tmp;
   }
 
+  ut_d(fil_page_type_validate(space, d));
   (*slot)->out_buf= d;
   return d;
 }
@@ -890,8 +897,7 @@ inline bool buf_page_t::flush(bool lru, fil_space_t *space)
                                               (write_frame ? write_frame
                                                : frame)));
       ut_ad(lsn >= oldest_modification());
-      if (lsn > log_sys.get_flushed_lsn())
-        log_write_up_to(lsn, true);
+      log_write_up_to(lsn, true);
     }
     space->io(IORequest{type, this, slot}, physical_offset(), size,
               write_frame, this);
@@ -995,59 +1001,52 @@ static page_id_t buf_flush_check_neighbors(const fil_space_t &space,
   return i;
 }
 
-MY_ATTRIBUTE((warn_unused_result))
-/** Apply freed_ranges to the file.
-@param writable whether the file is writable
-@return number of pages written or hole-punched */
-uint32_t fil_space_t::flush_freed(bool writable)
+MY_ATTRIBUTE((nonnull))
+/** Write punch-hole or zeroes of the freed ranges when
+innodb_immediate_scrub_data_uncompressed from the freed ranges.
+@param space   tablespace which may contain ranges of freed pages */
+static void buf_flush_freed_pages(fil_space_t *space)
 {
-  const bool punch_hole= chain.start->punch_hole == 1;
+  const bool punch_hole= space->chain.start->punch_hole == 1;
   if (!punch_hole && !srv_immediate_scrub_data_uncompressed)
-    return 0;
+    return;
+  lsn_t flush_to_disk_lsn= log_sys.get_flushed_lsn();
 
-  mysql_mutex_assert_not_owner(&buf_pool.flush_list_mutex);
-  mysql_mutex_assert_not_owner(&buf_pool.mutex);
-
-  freed_range_mutex.lock();
-  if (freed_ranges.empty() || log_sys.get_flushed_lsn() < get_last_freed_lsn())
+  std::unique_lock<std::mutex> freed_lock(space->freed_range_mutex);
+  if (space->freed_ranges.empty()
+      || flush_to_disk_lsn < space->get_last_freed_lsn())
   {
-    freed_range_mutex.unlock();
-    return 0;
+    freed_lock.unlock();
+    return;
   }
 
-  const unsigned physical{physical_size()};
+  range_set freed_ranges= std::move(space->freed_ranges);
+  freed_lock.unlock();
 
-  range_set freed= std::move(freed_ranges);
-  uint32_t written= 0;
-
-  if (!writable);
-  else if (punch_hole)
+  for (const auto &range : freed_ranges)
   {
-    for (const auto &range : freed)
+    const ulint physical_size= space->physical_size();
+
+    if (punch_hole)
     {
-      written+= range.last - range.first + 1;
-      reacquire();
-      io(IORequest(IORequest::PUNCH_RANGE),
-         os_offset_t{range.first} * physical,
-         (range.last - range.first + 1) * physical, nullptr);
+      space->reacquire();
+      space->io(IORequest(IORequest::PUNCH_RANGE),
+                          os_offset_t{range.first} * physical_size,
+                          (range.last - range.first + 1) * physical_size,
+                          nullptr);
     }
-  }
-  else
-  {
-    for (const auto &range : freed)
+    else
     {
-      written+= range.last - range.first + 1;
       for (os_offset_t i= range.first; i <= range.last; i++)
       {
-        reacquire();
-        io(IORequest(IORequest::WRITE_ASYNC), i * physical, physical,
-           const_cast<byte*>(field_ref_zero));
+        space->reacquire();
+        space->io(IORequest(IORequest::WRITE_ASYNC),
+                  i * physical_size, physical_size,
+                  const_cast<byte*>(field_ref_zero));
       }
     }
+    buf_pool.stat.n_pages_written+= (range.last - range.first + 1);
   }
-
-  freed_range_mutex.unlock();
-  return written;
 }
 
 /** Flushes to disk all flushable pages within the flush area
@@ -1170,12 +1169,14 @@ static ulint buf_free_from_unzip_LRU_list_batch(ulint max)
 
 /** Start writing out pages for a tablespace.
 @param id   tablespace identifier
-@return tablespace and number of pages written */
-static std::pair<fil_space_t*, uint32_t> buf_flush_space(const uint32_t id)
+@return tablespace
+@retval nullptr if the pages for this tablespace should be discarded */
+static fil_space_t *buf_flush_space(const uint32_t id)
 {
-  if (fil_space_t *space= fil_space_t::get(id))
-    return {space, space->flush_freed(true)};
-  return {nullptr, 0};
+  fil_space_t *space= fil_space_t::get(id);
+  if (space)
+    buf_flush_freed_pages(space);
+  return space;
 }
 
 struct flush_counters_t
@@ -1219,7 +1220,7 @@ static void buf_flush_LRU_list_batch(ulint max, flush_counters_t *n)
   ulint free_limit= srv_LRU_scan_depth;
 
   mysql_mutex_assert_owner(&buf_pool.mutex);
-  if (buf_pool.withdraw_target && buf_pool.is_shrinking())
+  if (buf_pool.withdraw_target && buf_pool.curr_size < buf_pool.old_size)
     free_limit+= buf_pool.withdraw_target - UT_LIST_GET_LEN(buf_pool.withdraw);
 
   const auto neighbors= UT_LIST_GET_LEN(buf_pool.LRU) < BUF_LRU_OLD_MIN_LEN
@@ -1261,17 +1262,10 @@ static void buf_flush_LRU_list_batch(ulint max, flush_counters_t *n)
       {
         if (last_space_id != space_id)
         {
-          buf_pool.lru_hp.set(bpage);
-          mysql_mutex_unlock(&buf_pool.mutex);
           if (space)
             space->release();
-          auto p= buf_flush_space(space_id);
-          space= p.first;
+          space= buf_flush_space(space_id);
           last_space_id= space_id;
-          mysql_mutex_lock(&buf_pool.mutex);
-          if (p.second)
-            buf_pool.stat.n_pages_written+= p.second;
-          goto retry;
         }
         else
           ut_ad(!space);
@@ -1302,7 +1296,6 @@ reacquire_mutex:
     must_skip:
       /* Can't evict or dispatch this block. Go to previous. */
       ut_ad(buf_pool.lru_hp.is_hp(prev));
-  retry:
     bpage= buf_pool.lru_hp.get();
   }
 
@@ -1407,28 +1400,10 @@ static ulint buf_do_flush_list_batch(ulint max_n, lsn_t lsn)
     {
       if (last_space_id != space_id)
       {
-        mysql_mutex_lock(&buf_pool.flush_list_mutex);
-        buf_pool.flush_hp.set(bpage);
-        mysql_mutex_unlock(&buf_pool.flush_list_mutex);
-        mysql_mutex_unlock(&buf_pool.mutex);
         if (space)
           space->release();
-        auto p= buf_flush_space(space_id);
-        space= p.first;
+        space= buf_flush_space(space_id);
         last_space_id= space_id;
-        mysql_mutex_lock(&buf_pool.mutex);
-        if (p.second)
-          buf_pool.stat.n_pages_written+= p.second;
-        mysql_mutex_lock(&buf_pool.flush_list_mutex);
-        bpage= buf_pool.flush_hp.get();
-        if (!bpage)
-          break;
-        if (bpage->id() != page_id)
-          continue;
-        buf_pool.flush_hp.set(UT_LIST_GET_PREV(list, bpage));
-        if (bpage->oldest_modification() <= 1 || !bpage->ready_for_flush())
-          goto next;
-        mysql_mutex_unlock(&buf_pool.flush_list_mutex);
       }
       else
         ut_ad(!space);
@@ -1456,7 +1431,6 @@ static ulint buf_do_flush_list_batch(ulint max_n, lsn_t lsn)
     }
 
     mysql_mutex_lock(&buf_pool.flush_list_mutex);
-  next:
     bpage= buf_pool.flush_hp.get();
   }
 
@@ -1553,15 +1527,12 @@ bool buf_flush_list_space(fil_space_t *space, ulint *n_flushed)
   bool may_have_skipped= false;
   ulint max_n_flush= srv_io_capacity;
 
-  bool acquired= space->acquire();
-  {
-    const uint32_t written{space->flush_freed(acquired)};
-    mysql_mutex_lock(&buf_pool.mutex);
-    if (written)
-      buf_pool.stat.n_pages_written+= written;
-  }
+  mysql_mutex_lock(&buf_pool.mutex);
   mysql_mutex_lock(&buf_pool.flush_list_mutex);
 
+  bool acquired= space->acquire();
+  buf_flush_freed_pages(space);
+
   for (buf_page_t *bpage= UT_LIST_GET_LAST(buf_pool.flush_list); bpage; )
   {
     ut_ad(bpage->oldest_modification());
@@ -1691,6 +1662,59 @@ ulint buf_flush_LRU(ulint max_n)
   return n_flushed;
 }
 
+#ifdef HAVE_PMEM
+# include <libpmem.h>
+#endif
+
+/** Write checkpoint information to the log header and release mutex.
+@param end_lsn    start LSN of the FILE_CHECKPOINT mini-transaction */
+inline void log_t::write_checkpoint(lsn_t end_lsn) noexcept
+{
+  ut_ad(!srv_read_only_mode);
+  ut_ad(end_lsn >= next_checkpoint_lsn);
+  ut_ad(end_lsn <= get_lsn());
+  ut_ad(end_lsn + SIZE_OF_FILE_CHECKPOINT <= get_lsn() ||
+        srv_shutdown_state > SRV_SHUTDOWN_INITIATED);
+
+  DBUG_PRINT("ib_log",
+             ("checkpoint at " LSN_PF " written", next_checkpoint_lsn));
+
+  auto n= next_checkpoint_no;
+  const size_t offset{(n & 1) ? CHECKPOINT_2 : CHECKPOINT_1};
+  static_assert(CPU_LEVEL1_DCACHE_LINESIZE >= 64, "efficiency");
+  static_assert(CPU_LEVEL1_DCACHE_LINESIZE <= 4096, "compatibility");
+  byte* c= my_assume_aligned<CPU_LEVEL1_DCACHE_LINESIZE>
+    (is_pmem() ? buf + offset : checkpoint_buf);
+  memset_aligned<CPU_LEVEL1_DCACHE_LINESIZE>(c, 0, CPU_LEVEL1_DCACHE_LINESIZE);
+  mach_write_to_8(my_assume_aligned<8>(c), next_checkpoint_lsn);
+  mach_write_to_8(my_assume_aligned<8>(c + 8), end_lsn);
+  mach_write_to_4(my_assume_aligned<4>(c + 60), my_crc32c(0, c, 60));
+#ifdef HAVE_PMEM
+  if (is_pmem())
+    pmem_persist(c, 64);
+  else
+#endif
+  {
+    n_pending_checkpoint_writes++;
+    mysql_mutex_unlock(&mutex);
+    /* FIXME: issue an asynchronous write */
+    log.write(offset, {c, get_block_size()});
+    if (srv_file_flush_method != SRV_O_DSYNC)
+      ut_a(log.flush());
+    mysql_mutex_lock(&mutex);
+    n_pending_checkpoint_writes--;
+  }
+
+  ut_ad(!n_pending_checkpoint_writes);
+  next_checkpoint_no++;
+  last_checkpoint_lsn= next_checkpoint_lsn;
+
+  DBUG_PRINT("ib_log", ("checkpoint ended at " LSN_PF ", flushed to " LSN_PF,
+                        next_checkpoint_lsn, get_flushed_lsn()));
+
+  mysql_mutex_unlock(&mutex);
+}
+
 /** Initiate a log checkpoint, discarding the start of the log.
 @param oldest_lsn   the checkpoint LSN
 @param end_lsn      log_sys.get_lsn()
@@ -1701,31 +1725,21 @@ static bool log_checkpoint_low(lsn_t oldest_lsn, lsn_t end_lsn)
   mysql_mutex_assert_owner(&log_sys.mutex);
   ut_ad(oldest_lsn <= end_lsn);
   ut_ad(end_lsn == log_sys.get_lsn());
+  ut_ad(!recv_no_log_write);
 
-  ut_ad(oldest_lsn >= log_sys.last_checkpoint_lsn);
-  const lsn_t age= oldest_lsn - log_sys.last_checkpoint_lsn;
-
-  if (age > SIZE_OF_FILE_CHECKPOINT + log_sys.framing_size())
-    /* Some log has been written since the previous checkpoint. */;
-  else if (age > SIZE_OF_FILE_CHECKPOINT &&
-           !((log_sys.log.calc_lsn_offset(oldest_lsn) ^
-              log_sys.log.calc_lsn_offset(log_sys.last_checkpoint_lsn)) &
-             ~lsn_t{OS_FILE_LOG_BLOCK_SIZE - 1}))
-    /* Some log has been written to the same log block. */;
-  else if (srv_shutdown_state > SRV_SHUTDOWN_INITIATED)
-    /* MariaDB startup expects the redo log file to be logically empty
-    (not even containing a FILE_CHECKPOINT record) after a clean shutdown.
-    Perform an extra checkpoint at shutdown. */;
-  else
+  if (oldest_lsn == log_sys.last_checkpoint_lsn ||
+      (oldest_lsn == end_lsn && oldest_lsn == log_sys.last_checkpoint_lsn +
+       (log_sys.is_encrypted()
+        ? SIZE_OF_FILE_CHECKPOINT + 8 : SIZE_OF_FILE_CHECKPOINT)))
   {
     /* Do nothing, because nothing was logged (other than a
     FILE_CHECKPOINT record) since the previous checkpoint. */
+  do_nothing:
     mysql_mutex_unlock(&log_sys.mutex);
     return true;
   }
 
-  ut_ad(!recv_no_log_write);
-
+  ut_ad(oldest_lsn > log_sys.last_checkpoint_lsn);
   /* Repeat the FILE_MODIFY records after the checkpoint, in case some
   log records between the checkpoint and log_sys.lsn need them.
   Finally, write a FILE_CHECKPOINT record. Redo log apply expects to
@@ -1736,28 +1750,17 @@ static bool log_checkpoint_low(lsn_t oldest_lsn, lsn_t end_lsn)
   dirty pages are flushed to the tablespace files.  At this point,
   because we hold log_sys.mutex, mtr_t::commit() in other threads will
   be blocked, and no pages can be added to the flush lists. */
-  lsn_t flush_lsn= oldest_lsn;
-
-  if (fil_names_clear(flush_lsn, oldest_lsn != end_lsn ||
-                      srv_shutdown_state <= SRV_SHUTDOWN_INITIATED))
-  {
-    flush_lsn= log_sys.get_lsn();
-    ut_ad(flush_lsn >= end_lsn + SIZE_OF_FILE_CHECKPOINT);
-    mysql_mutex_unlock(&log_sys.mutex);
-    log_write_up_to(flush_lsn, true, true);
-    mysql_mutex_lock(&log_sys.mutex);
-    if (log_sys.last_checkpoint_lsn >= oldest_lsn)
-    {
-      mysql_mutex_unlock(&log_sys.mutex);
-      return true;
-    }
-  }
-  else
-    ut_ad(oldest_lsn >= log_sys.last_checkpoint_lsn);
+  const lsn_t flush_lsn{fil_names_clear(oldest_lsn)};
+  ut_ad(flush_lsn >= end_lsn + SIZE_OF_FILE_CHECKPOINT);
+  mysql_mutex_unlock(&log_sys.mutex);
+  log_write_up_to(flush_lsn, true);
+  mysql_mutex_lock(&log_sys.mutex);
+  if (log_sys.last_checkpoint_lsn >= oldest_lsn)
+    goto do_nothing;
 
   ut_ad(log_sys.get_flushed_lsn() >= flush_lsn);
 
-  if (log_sys.checkpoint_pending)
+  if (log_sys.n_pending_checkpoint_writes)
   {
     /* A checkpoint write is running */
     mysql_mutex_unlock(&log_sys.mutex);
@@ -1765,7 +1768,7 @@ static bool log_checkpoint_low(lsn_t oldest_lsn, lsn_t end_lsn)
   }
 
   log_sys.next_checkpoint_lsn= oldest_lsn;
-  log_write_checkpoint_info(end_lsn);
+  log_sys.write_checkpoint(end_lsn);
   mysql_mutex_assert_not_owner(&log_sys.mutex);
 
   return true;
@@ -1795,8 +1798,8 @@ static bool log_checkpoint()
   mysql_mutex_lock(&log_sys.flush_order_mutex);
   mysql_mutex_lock(&buf_pool.flush_list_mutex);
   const lsn_t oldest_lsn= buf_pool.get_oldest_modification(end_lsn);
-  mysql_mutex_unlock(&buf_pool.flush_list_mutex);
   mysql_mutex_unlock(&log_sys.flush_order_mutex);
+  mysql_mutex_unlock(&buf_pool.flush_list_mutex);
   return log_checkpoint_low(oldest_lsn, end_lsn);
 }
 
@@ -1880,8 +1883,7 @@ ATTRIBUTE_COLD void buf_flush_wait_flushed(lsn_t sync_lsn)
     to happen until now. There could be an outstanding FILE_CHECKPOINT
     record from a previous fil_names_clear() call, which we must
     write out before we can advance the checkpoint. */
-    if (sync_lsn > log_sys.get_flushed_lsn())
-      log_write_up_to(sync_lsn, true);
+    log_write_up_to(sync_lsn, true);
     log_checkpoint();
   }
 }
@@ -1954,8 +1956,8 @@ ATTRIBUTE_COLD static void buf_flush_sync_for_checkpoint(lsn_t lsn)
     const lsn_t newest_lsn= log_sys.get_lsn();
     mysql_mutex_lock(&log_sys.flush_order_mutex);
     mysql_mutex_lock(&buf_pool.flush_list_mutex);
-    lsn_t measure= buf_pool.get_oldest_modification(0);
     mysql_mutex_unlock(&log_sys.flush_order_mutex);
+    lsn_t measure= buf_pool.get_oldest_modification(0);
     const lsn_t checkpoint_lsn= measure ? measure : newest_lsn;
 
     if (!recv_recovery_is_on() &&
@@ -2474,7 +2476,11 @@ NOTE: The calling thread is not allowed to hold any buffer page latches! */
 void buf_flush_sync()
 {
   if (recv_recovery_is_on())
+  {
+    mysql_mutex_lock(&recv_sys.mutex);
     recv_sys.apply(true);
+    mysql_mutex_unlock(&recv_sys.mutex);
+  }
 
   thd_wait_begin(nullptr, THD_WAIT_DISKIO);
   tpool::tpool_wait_begin();
