@@ -1,7 +1,7 @@
 /*****************************************************************************
 
 Copyright (c) 1996, 2016, Oracle and/or its affiliates. All Rights Reserved.
-Copyright (c) 2017, 2022, MariaDB Corporation.
+Copyright (c) 2017, 2021, MariaDB Corporation.
 
 This program is free software; you can redistribute it and/or modify it under
 the terms of the GNU General Public License as published by the Free Software
@@ -536,15 +536,15 @@ dict_create_sys_fields_tuple(
 	dict_field_t*	field;
 	dfield_t*	dfield;
 	byte*		ptr;
-	ibool		index_contains_column_prefix_field	= FALSE;
-	ulint		j;
+	bool		wide_pos = false;
 
 	ut_ad(index);
 	ut_ad(heap);
 
-	for (j = 0; j < index->n_fields; j++) {
-		if (dict_index_get_nth_field(index, j)->prefix_len > 0) {
-			index_contains_column_prefix_field = TRUE;
+	for (unsigned j = 0; j < index->n_fields; j++) {
+		const dict_field_t* f = dict_index_get_nth_field(index, j);
+		if (f->prefix_len || f->descending) {
+			wide_pos = true;
 			break;
 		}
 	}
@@ -569,12 +569,15 @@ dict_create_sys_fields_tuple(
 
 	ptr = static_cast<byte*>(mem_heap_alloc(heap, 4));
 
-	if (index_contains_column_prefix_field) {
-		/* If there are column prefix fields in the index, then
-		we store the number of the field to the 2 HIGH bytes
-		and the prefix length to the 2 low bytes, */
-
-		mach_write_to_4(ptr, (fld_no << 16) + field->prefix_len);
+	if (wide_pos) {
+		/* If there are column prefixes or columns with
+		descending order in the index, then we write the
+		field number to the 16 most significant bits,
+		the DESC flag to bit 15, and the prefix length
+		in the 15 least significant bits. */
+		mach_write_to_4(ptr, (fld_no << 16)
+				| (!!field->descending) << 15
+				| field->prefix_len);
 	} else {
 		/* Else we store the number of the field to the 2 LOW bytes.
 		This is to keep the storage format compatible with
@@ -751,32 +754,28 @@ dict_create_index_tree_step(
 	mtr.start();
 
 	search_tuple = dict_create_search_tuple(node->ind_row, node->heap);
-	node->page_no = FIL_NULL;
 
-	dberr_t err =
-		btr_pcur_open(UT_LIST_GET_FIRST(dict_sys.sys_indexes->indexes),
-			      search_tuple, PAGE_CUR_L, BTR_MODIFY_LEAF,
-			      &pcur, &mtr);
-
-	if (err != DB_SUCCESS) {
-func_exit:
-		mtr.commit();
-		return err;
-	}
+	btr_pcur_open(UT_LIST_GET_FIRST(dict_sys.sys_indexes->indexes),
+		      search_tuple, PAGE_CUR_L, BTR_MODIFY_LEAF,
+		      &pcur, &mtr);
 
 	btr_pcur_move_to_next_user_rec(&pcur, &mtr);
 
-	if (UNIV_UNLIKELY(btr_pcur_is_after_last_on_page(&pcur))) {
-		err = DB_CORRUPTION;
-		goto func_exit;
-	}
 
-	if (index->is_readable()) {
+	dberr_t		err = DB_SUCCESS;
+
+	if (!index->is_readable()) {
+		node->page_no = FIL_NULL;
+	} else {
 		index->set_modified(mtr);
 
 		node->page_no = btr_create(
 			index->type, index->table->space,
-			index->id, index, &mtr, &err);
+			index->id, index, &mtr);
+
+		if (node->page_no == FIL_NULL) {
+			err = DB_OUT_OF_FILE_SPACE;
+		}
 
 		DBUG_EXECUTE_IF("ib_import_create_index_failure_1",
 				node->page_no = FIL_NULL;
@@ -790,13 +789,16 @@ func_exit:
 	ut_ad(len == 4);
 	mtr.write<4,mtr_t::MAYBE_NOP>(*btr_pcur_get_block(&pcur), data,
 				      node->page_no);
-	goto func_exit;
+
+	mtr.commit();
+
+	return(err);
 }
 
 /***************************************************************//**
 Creates an index tree for the index if it is not a member of a cluster.
 Don't update SYSTEM TABLES.
-@return error code */
+@return DB_SUCCESS or DB_OUT_OF_FILE_SPACE */
 dberr_t
 dict_create_index_tree_in_mem(
 /*==========================*/
@@ -816,14 +818,13 @@ dict_create_index_tree_in_mem(
 	ut_ad(index->is_readable());
 	ut_ad(!(index->table->flags2 & DICT_TF2_DISCARDED));
 
-	dberr_t err;
 	index->page = btr_create(index->type, index->table->space,
-				 index->id, index, &mtr, &err);
+				 index->id, index, &mtr);
 	mtr_commit(&mtr);
 
 	index->trx_id = trx->id;
 
-	return err;
+	return index->page == FIL_NULL ? DB_OUT_OF_FILE_SPACE : DB_SUCCESS;
 }
 
 /** Drop the index tree associated with a row in SYS_INDEXES table.
@@ -848,7 +849,7 @@ uint32_t dict_drop_index_tree(btr_pcur_t *pcur, trx_t *trx, mtr_t *mtr)
       len > DICT_NUM_FIELDS__SYS_INDEXES)
   {
 rec_corrupted:
-    sql_print_error("InnoDB: Corrupted SYS_INDEXES record");
+    ib::error() << "Corrupted SYS_INDEXES record";
     return 0;
   }
 
@@ -1107,7 +1108,15 @@ dict_create_table_step(
 function_exit:
 	trx->error_state = err;
 
-	if (err != DB_SUCCESS) {
+	if (err == DB_SUCCESS) {
+		/* Ok: do nothing */
+
+	} else if (err == DB_LOCK_WAIT) {
+
+		return(NULL);
+	} else {
+		/* SQL error detected */
+
 		return(NULL);
 	}
 
@@ -1291,8 +1300,16 @@ dict_create_index_step(
 function_exit:
 	trx->error_state = err;
 
-	if (UNIV_UNLIKELY(err != DB_SUCCESS)) {
-		return nullptr;
+	if (err == DB_SUCCESS) {
+		/* Ok: do nothing */
+
+	} else if (err == DB_LOCK_WAIT) {
+
+		return(NULL);
+	} else {
+		/* SQL error detected */
+
+		return(NULL);
 	}
 
 	thr->run_node = que_node_get_parent(node);
@@ -1314,7 +1331,7 @@ bool dict_sys_t::load_sys_tables()
   {
     sys_foreign= nullptr;
     mismatch= true;
-    sql_print_error("InnoDB: Invalid definition of SYS_FOREIGN");
+    ib::error() << "Invalid definition of SYS_FOREIGN";
   }
   if (!(sys_foreign_cols= load_table(SYS_TABLE[SYS_FOREIGN_COLS],
                                      DICT_ERR_IGNORE_FK_NOKEY)));
@@ -1326,7 +1343,7 @@ bool dict_sys_t::load_sys_tables()
   {
     sys_foreign_cols= nullptr;
     mismatch= true;
-    sql_print_error("InnoDB: Invalid definition of SYS_FOREIGN_COLS");
+    ib::error() << "Invalid definition of SYS_FOREIGN_COLS";
   }
   if (!(sys_virtual= load_table(SYS_TABLE[SYS_VIRTUAL],
                                 DICT_ERR_IGNORE_FK_NOKEY)));
@@ -1337,7 +1354,7 @@ bool dict_sys_t::load_sys_tables()
   {
     sys_virtual= nullptr;
     mismatch= true;
-    sql_print_error("InnoDB: Invalid definition of SYS_VIRTUAL");
+    ib::error() << "Invalid definition of SYS_VIRTUAL";
   }
   unlock();
   return mismatch;
@@ -1353,8 +1370,8 @@ dberr_t dict_sys_t::create_or_check_sys_tables()
 
   if (load_sys_tables())
   {
-    sql_print_information("InnoDB: Set innodb_read_only=1 "
-                          "or innodb_force_recovery=3 to start up");
+    ib::info() << "Set innodb_read_only=1 or innodb_force_recovery=3"
+                  " to start up";
     return DB_CORRUPTION;
   }
 
@@ -1386,7 +1403,7 @@ dberr_t dict_sys_t::create_or_check_sys_tables()
   const auto srv_file_per_table_backup= srv_file_per_table;
   srv_file_per_table= 0;
   dberr_t error;
-  span<const char> tablename;
+  const char *tablename;
 
   if (!sys_foreign)
   {
@@ -1404,11 +1421,9 @@ dberr_t dict_sys_t::create_or_check_sys_tables()
                         "END;\n", trx);
     if (UNIV_UNLIKELY(error != DB_SUCCESS))
     {
-      tablename= SYS_TABLE[SYS_FOREIGN];
+      tablename= SYS_TABLE[SYS_FOREIGN].data();
 err_exit:
-      sql_print_error("InnoDB: Creation of %.*s failed: %s",
-                      int(tablename.size()), tablename.data(),
-                      ut_strerr(error));
+      ib::error() << "Creation of " << tablename << " failed: " << error;
       trx->rollback();
       row_mysql_unlock_data_dictionary(trx);
       trx->free();
@@ -1428,7 +1443,7 @@ err_exit:
                         "END;\n", trx);
     if (UNIV_UNLIKELY(error != DB_SUCCESS))
     {
-      tablename= SYS_TABLE[SYS_FOREIGN_COLS];
+      tablename= SYS_TABLE[SYS_FOREIGN_COLS].data();
       goto err_exit;
     }
   }
@@ -1443,7 +1458,7 @@ err_exit:
                         "END;\n", trx);
     if (UNIV_UNLIKELY(error != DB_SUCCESS))
     {
-      tablename= SYS_TABLE[SYS_VIRTUAL];
+      tablename= SYS_TABLE[SYS_VIRTUAL].data();
       goto err_exit;
     }
   }
@@ -1457,11 +1472,10 @@ err_exit:
   if (sys_foreign);
   else if (!(sys_foreign= load_table(SYS_TABLE[SYS_FOREIGN])))
   {
-    tablename= SYS_TABLE[SYS_FOREIGN];
+    tablename= SYS_TABLE[SYS_FOREIGN].data();
 load_fail:
     unlock();
-    sql_print_error("InnoDB: Failed to CREATE TABLE %.*s",
-                    int(tablename.size()), tablename.data());
+    ib::error() << "Failed to CREATE TABLE " << tablename;
     return DB_TABLE_NOT_FOUND;
   }
   else
@@ -1470,7 +1484,7 @@ load_fail:
   if (sys_foreign_cols);
   else if (!(sys_foreign_cols= load_table(SYS_TABLE[SYS_FOREIGN_COLS])))
   {
-    tablename= SYS_TABLE[SYS_FOREIGN_COLS];
+    tablename= SYS_TABLE[SYS_FOREIGN_COLS].data();
     goto load_fail;
   }
   else
@@ -1479,7 +1493,7 @@ load_fail:
   if (sys_virtual);
   else if (!(sys_virtual= load_table(SYS_TABLE[SYS_VIRTUAL])))
   {
-    tablename= SYS_TABLE[SYS_VIRTUAL];
+    tablename= SYS_TABLE[SYS_VIRTUAL].data();
     goto load_fail;
   }
   else
@@ -1502,14 +1516,12 @@ dict_foreign_eval_sql(
 	const char*	id,	/*!< in: foreign key id */
 	trx_t*		trx)	/*!< in/out: transaction */
 {
+	dberr_t	error;
 	FILE*	ef	= dict_foreign_err_file;
 
-	dberr_t error = que_eval_sql(info, sql, trx);
+	error = que_eval_sql(info, sql, trx);
 
-	switch (error) {
-	case DB_SUCCESS:
-		break;
-	case DB_DUPLICATE_KEY:
+	if (error == DB_DUPLICATE_KEY) {
 		mysql_mutex_lock(&dict_foreign_err_mutex);
 		rewind(ef);
 		ut_print_timestamp(ef);
@@ -1529,11 +1541,15 @@ dict_foreign_eval_sql(
 		      "names can occur. Workaround: name your constraints\n"
 		      "explicitly with unique names.\n",
 		      ef);
-		goto release;
-	default:
-		sql_print_error("InnoDB: "
-				"Foreign key constraint creation failed: %s",
-				ut_strerr(error));
+
+		mysql_mutex_unlock(&dict_foreign_err_mutex);
+
+		return(error);
+	}
+
+	if (UNIV_UNLIKELY(error != DB_SUCCESS)) {
+		ib::error() << "Foreign key constraint creation failed: "
+			<< error;
 
 		mysql_mutex_lock(&dict_foreign_err_mutex);
 		ut_print_timestamp(ef);
@@ -1543,11 +1559,12 @@ dict_foreign_eval_sql(
 		fputs(".\n"
 		      "See the MariaDB .err log in the datadir"
 		      " for more information.\n", ef);
-release:
 		mysql_mutex_unlock(&dict_foreign_err_mutex);
+
+		return(error);
 	}
 
-	return error;
+	return(DB_SUCCESS);
 }
 
 /********************************************************************//**
