@@ -1,7 +1,7 @@
 /*****************************************************************************
 
 Copyright (c) 1995, 2015, Oracle and/or its affiliates. All rights reserved.
-Copyright (c) 2019, 2022, MariaDB Corporation.
+Copyright (c) 2019, 2021, MariaDB Corporation.
 
 This program is free software; you can redistribute it and/or modify it under
 the terms of the GNU General Public License as published by the Free Software
@@ -89,25 +89,20 @@ this must be equal to srv_page_size */
 class page_id_t
 {
 public:
+
   /** Constructor from (space, page_no).
   @param space	 tablespace id
   @param page_no page number */
-  constexpr page_id_t(uint32_t space, uint32_t page_no) :
+  page_id_t(uint32_t space, uint32_t page_no) :
     m_id(uint64_t{space} << 32 | page_no) {}
 
-  constexpr page_id_t(uint64_t id) : m_id(id) {}
-  constexpr bool operator==(const page_id_t& rhs) const
-  { return m_id == rhs.m_id; }
-  constexpr bool operator!=(const page_id_t& rhs) const
-  { return m_id != rhs.m_id; }
-  constexpr bool operator<(const page_id_t& rhs) const
-  { return m_id < rhs.m_id; }
-  constexpr bool operator>(const page_id_t& rhs) const
-  { return m_id > rhs.m_id; }
-  constexpr bool operator<=(const page_id_t& rhs) const
-  { return m_id <= rhs.m_id; }
-  constexpr bool operator>=(const page_id_t& rhs) const
-  { return m_id >= rhs.m_id; }
+  page_id_t(uint64_t id) : m_id(id) {}
+  bool operator==(const page_id_t& rhs) const { return m_id == rhs.m_id; }
+  bool operator!=(const page_id_t& rhs) const { return m_id != rhs.m_id; }
+  bool operator<(const page_id_t& rhs) const { return m_id < rhs.m_id; }
+  bool operator>(const page_id_t& rhs) const { return m_id > rhs.m_id; }
+  bool operator<=(const page_id_t& rhs) const { return m_id <= rhs.m_id; }
+  bool operator>=(const page_id_t& rhs) const { return m_id >= rhs.m_id; }
   page_id_t &operator--() { ut_ad(page_no()); m_id--; return *this; }
   page_id_t &operator++()
   {
@@ -128,16 +123,15 @@ public:
 
   /** Retrieve the tablespace id.
   @return tablespace id */
-  constexpr uint32_t space() const { return static_cast<uint32_t>(m_id >> 32); }
+  uint32_t space() const { return static_cast<uint32_t>(m_id >> 32); }
 
   /** Retrieve the page number.
   @return page number */
-  constexpr uint32_t page_no() const { return static_cast<uint32_t>(m_id); }
+  uint32_t page_no() const { return static_cast<uint32_t>(m_id); }
 
   /** Retrieve the fold value.
   @return fold value */
-  constexpr ulint fold() const
-  { return (ulint{space()} << 20) + space() + page_no(); }
+  ulint fold() const { return (space() << 20) + space() + page_no(); }
 
   /** Reset the page number only.
   @param[in]	page_no	page number */
@@ -146,7 +140,7 @@ public:
     m_id= (m_id & ~uint64_t{0} << 32) | page_no;
   }
 
-  constexpr ulonglong raw() const { return m_id; }
+  ulonglong raw() { return m_id; }
 
 private:
   /** The page identifier */
