@@ -24,7 +24,8 @@ Mini-transaction buffer
 Created 11/26/1995 Heikki Tuuri
 *******************************************************/
 
-#pragma once
+#ifndef mtr0mtr_h
+#define mtr0mtr_h
 
 #include "fil0fil.h"
 #include "dyn0buf.h"
@@ -53,6 +54,16 @@ savepoint. */
 #define mtr_memo_release(m, o, t)					\
 				(m)->memo_release((o), (t))
 
+/** Print info of an mtr handle. */
+#define mtr_print(m)		(m)->print()
+
+/** Return the log object of a mini-transaction buffer.
+@return	log */
+#define mtr_get_log(m)		(m)->get_log()
+
+/** Push an object to an mtr memo stack. */
+#define mtr_memo_push(m, o, t)	(m)->memo_push(o, t)
+
 #ifdef UNIV_PFS_RWLOCK
 # define mtr_s_lock_index(i,m)	(m)->s_lock(__FILE__, __LINE__, &(i)->lock)
 # define mtr_x_lock_index(i,m)	(m)->x_lock(__FILE__, __LINE__, &(i)->lock)
@@ -89,32 +100,18 @@ struct mtr_t {
   /** Commit the mini-transaction. */
   void commit();
 
-  /** Release latches till savepoint. To simplify the code only
-  MTR_MEMO_S_LOCK and MTR_MEMO_PAGE_S_FIX slot types are allowed to be
-  released, otherwise it would be neccesary to add one more argument in the
-  function to point out what slot types are allowed for rollback, and this
-  would be overengineering as currently the function is used only in one place
-  in the code.
-  @param savepoint   savepoint, can be obtained with get_savepoint */
-  void rollback_to_savepoint(ulint savepoint);
-
   /** Commit a mini-transaction that is shrinking a tablespace.
   @param space   tablespace that is being shrunk */
   ATTRIBUTE_COLD void commit_shrink(fil_space_t &space);
 
-  /** Commit a mini-transaction that is deleting or renaming a file.
-  @param space   tablespace that is being renamed or deleted
-  @param name    new file name (nullptr=the file will be deleted)
-  @return whether the operation succeeded */
-  ATTRIBUTE_COLD bool commit_file(fil_space_t &space, const char *name);
-
   /** Commit a mini-transaction that did not modify any pages,
   but generated some redo log on a higher level, such as
   FILE_MODIFY records and an optional FILE_CHECKPOINT marker.
   The caller must hold log_sys.mutex.
   This is to be used at log_checkpoint().
-  @param checkpoint_lsn   the log sequence number of a checkpoint, or 0 */
-  void commit_files(lsn_t checkpoint_lsn= 0);
+  @param checkpoint_lsn   the log sequence number of a checkpoint, or 0
+  @return current LSN */
+  lsn_t commit_files(lsn_t checkpoint_lsn= 0);
 
   /** @return mini-transaction savepoint (current size of m_memo) */
   ulint get_savepoint() const { ut_ad(is_active()); return m_memo.size(); }
@@ -142,18 +139,10 @@ struct mtr_t {
   mtr_log_t get_log_mode() const
   {
     static_assert(MTR_LOG_ALL == 0, "efficiency");
+    ut_ad(m_log_mode <= MTR_LOG_NO_REDO);
     return static_cast<mtr_log_t>(m_log_mode);
   }
 
-  /** @return whether log is to be written for changes */
-  bool is_logged() const
-  {
-    static_assert(MTR_LOG_ALL == 0, "efficiency");
-    static_assert(MTR_LOG_NONE & MTR_LOG_NO_REDO, "efficiency");
-    static_assert(!(MTR_LOG_NONE & MTR_LOG_SUB), "efficiency");
-    return !(m_log_mode & MTR_LOG_NONE);
-  }
-
   /** Change the logging mode.
   @param mode	 logging mode
   @return	old mode */
@@ -164,23 +153,10 @@ struct mtr_t {
     return old_mode;
   }
 
-  /** Set the log mode of a sub-minitransaction
-  @param mtr  parent mini-transaction */
-  void set_log_mode_sub(const mtr_t &mtr)
-  {
-    ut_ad(mtr.m_log_mode == MTR_LOG_ALL || mtr.m_log_mode == MTR_LOG_NO_REDO);
-    m_log_mode= mtr.m_log_mode | MTR_LOG_SUB;
-    static_assert((MTR_LOG_SUB | MTR_LOG_NO_REDO) == MTR_LOG_NO_REDO, "");
-  }
-
   /** Check if we are holding a block latch in exclusive mode
   @param block  buffer pool block to search for */
   bool have_x_latch(const buf_block_t &block) const;
 
-  /** Check if we are holding a block latch in S or U mode
-  @param block  buffer pool block to search for */
-  bool have_u_or_x_latch(const buf_block_t &block) const;
-
 	/** Copy the tablespaces associated with the mini-transaction
 	(needed for generating FILE_MODIFY records)
 	@param[in]	mtr	mini-transaction that may modify
@@ -300,9 +276,9 @@ struct mtr_t {
   @param[in]	type	object type: MTR_MEMO_PAGE_X_FIX, ... */
   void release_page(const void *ptr, mtr_memo_type_t type);
 
+private:
   /** Note that the mini-transaction will modify data. */
   void flag_modified() { m_modifications = true; }
-private:
   /** Mark the given latched page as modified.
   @param block   page that will be modified */
   void modify(const buf_block_t& block);
@@ -340,15 +316,6 @@ public:
   @param rw_latch RW_S_LATCH, RW_SX_LATCH, RW_X_LATCH, RW_NO_LATCH */
   void page_lock(buf_block_t *block, ulint rw_latch);
 
-  /** Register a page latch on a buffer-fixed block was buffer-fixed.
-  @param latch   latch type */
-  void u_lock_register(ulint savepoint)
-  {
-    mtr_memo_slot_t *slot= m_memo.at<mtr_memo_slot_t*>(savepoint);
-    ut_ad(slot->type == MTR_MEMO_BUF_FIX);
-    slot->type= MTR_MEMO_PAGE_SX_FIX;
-  }
-
   /** Upgrade U locks on a block to X */
   void page_lock_upgrade(const buf_block_t &block);
   /** Upgrade X lock to X */
@@ -385,13 +352,30 @@ public:
 		const byte*	ptr,
 		ulint		flags) const;
 
+	/** Print info of an mtr handle. */
+	void print() const;
+
 	/** @return true if mini-transaction contains modifications. */
 	bool has_modifications() const { return m_modifications; }
+
+	/** @return the memo stack */
+	const mtr_buf_t* get_memo() const { return &m_memo; }
+
+	/** @return the memo stack */
+	mtr_buf_t* get_memo() { return &m_memo; }
 #endif /* UNIV_DEBUG */
 
 	/** @return true if a record was added to the mini-transaction */
 	bool is_dirty() const { return m_made_dirty; }
 
+	/** Get the buffered redo log of this mini-transaction.
+	@return	redo log */
+	const mtr_buf_t* get_log() const { return &m_log; }
+
+	/** Get the buffered redo log of this mini-transaction.
+	@return	redo log */
+	mtr_buf_t* get_log() { return &m_log; }
+
 	/** Push an object to an mtr memo stack.
 	@param object	object
 	@param type	object type: MTR_MEMO_S_LOCK, ... */
@@ -403,14 +387,6 @@ public:
 	static inline bool is_block_dirtied(const buf_block_t* block)
 		MY_ATTRIBUTE((warn_unused_result));
 
-  /** @return the size of the log is empty */
-  size_t get_log_size() const { return m_log.size(); }
-  /** @return whether the log and memo are empty */
-  bool is_empty() const { return m_memo.size() == 0 && m_log.size() == 0; }
-
-  /** Write an OPT_PAGE_CHECKSUM record. */
-  inline void page_checksum(const buf_page_t &bpage);
-
   /** Write request types */
   enum write_type
   {
@@ -509,9 +485,9 @@ public:
   @param[in,out]        b       buffer page */
   void init(buf_block_t *b);
   /** Free a page.
-  @param space   tablespace
-  @param offset  offset of the page to be freed */
-  void free(const fil_space_t &space, uint32_t offset);
+  @param[in]      space   tablespace contains page to be freed
+  @param[in]      offset  page offset to be freed */
+  inline void free(fil_space_t &space, uint32_t offset);
   /** Write log for partly initializing a B-tree or R-tree page.
   @param block    B-tree or R-tree page
   @param comp     false=ROW_FORMAT=REDUNDANT, true=COMPACT or DYNAMIC */
@@ -648,6 +624,13 @@ private:
   @param type   extended record subtype; @see mrec_ext_t */
   inline void log_write_extended(const buf_block_t &block, byte type);
 
+  /** Write a FILE_MODIFY record when a non-predefined persistent
+  tablespace was modified for the first time since fil_names_clear(). */
+  ATTRIBUTE_NOINLINE ATTRIBUTE_COLD void name_write();
+
+  /** Encrypt the log */
+  ATTRIBUTE_NOINLINE void encrypt();
+
   /** Append the redo log records to the redo log buffer.
   @return {start_lsn,flush_ahead} */
   std::pair<lsn_t,page_flush_ahead> do_write();
@@ -655,7 +638,7 @@ private:
   /** Append the redo log records to the redo log buffer.
   @param len   number of bytes to write
   @return {start_lsn,flush_ahead} */
-  inline std::pair<lsn_t,page_flush_ahead> finish_write(ulint len);
+  std::pair<lsn_t,page_flush_ahead> finish_write(size_t len);
 
   /** Release the resources */
   inline void release_resources();
@@ -701,6 +684,9 @@ private:
   /** whether the pages has been trimmed */
   uint16_t m_trim_pages:1;
 
+  /** CRC-32C of m_log */
+  uint32_t m_crc;
+
 #ifdef UNIV_DEBUG
   /** Persistent user tablespace associated with the
   mini-transaction, or 0 (TRX_SYS_SPACE) if none yet */
@@ -726,3 +712,5 @@ private:
 };
 
 #include "mtr0mtr.inl"
+
+#endif /* mtr0mtr_h */
