@@ -1,7 +1,7 @@
 /*****************************************************************************
 
 Copyright (c) 1995, 2015, Oracle and/or its affiliates. All Rights Reserved.
-Copyright (c) 2017, 2020, MariaDB Corporation.
+Copyright (c) 2017, 2022, MariaDB Corporation.
 
 This program is free software; you can redistribute it and/or modify it under
 the terms of the GNU General Public License as published by the Free Software
@@ -24,14 +24,11 @@ Mini-transaction buffer global types
 Created 11/26/1995 Heikki Tuuri
 *******************************************************/
 
-#ifndef mtr0types_h
-#define mtr0types_h
+#pragma once
 
-#ifndef UNIV_INNOCHECKSUM
-#include "sync0rw.h"
-#else
-#include "univ.i"
-#endif /* UNIV_INNOCHECKSUM */
+#include "buf0types.h"
+
+#include "ut0byte.h"
 
 struct mtr_t;
 
@@ -50,17 +47,17 @@ enum mtr_log_t {
 
 /*
 A mini-transaction is a stream of records that is always terminated by
-a NUL byte. The first byte of a mini-transaction record is never NUL,
-but NUL bytes can occur within mini-transaction records. The first
-bytes of each record will explicitly encode the length of the record.
-NUL bytes also acts as padding in log blocks, that is, there can be
-multiple sucessive NUL bytes between mini-transactions in a redo log
-block.
+a byte 0x00 or 0x01. The first byte of a mini-transaction record is
+never one of these bytes, but these bytes can occur within mini-transaction
+records.
 
 The first byte of the record would contain a record type, flags, and a
 part of length. The optional second byte of the record will contain
 more length. (Not needed for short records.)
 
+For example, because the length of an INIT_PAGE record is 3 to 11 bytes,
+the first byte will be 0x02 to 0x0a, indicating the number of subsequent bytes.
+
 Bit 7 of the first byte of a redo log record is the same_page flag.
 If same_page=1, the record is referring to the same page as the
 previous record. Records that do not refer to data pages but to file
@@ -115,13 +112,9 @@ If same_page=1 is set in a record that follows a same_page=0 record
 in a mini-transaction, the tablespace identifier and page number
 fields will be omitted.
 
-(For some file-oriented records (if same_page=1 for the first records
-of a mini-transaction), we will write tablespace identifier using the
-same 1-to-5-byte encoding. TBD: describe the exact format of
-file-oriented records. With MDEV-14425, we could write file-level log
-records to a separate file, not interleaved with page-level redo log
-at all. We could reserve the file ib_logfile0 for checkpoint information
-and for file-level redo log records.)
+For FILE_ records (if same_page=1 for the first record
+of a mini-transaction), we will write a tablespace identifier and
+a page number (always 0) using the same 1-to-5-byte encoding.
 
 For FREE_PAGE or INIT_PAGE, if same_page=1, the record will be treated
 as corrupted (or reserved for future extension).  The type code must
@@ -193,8 +186,11 @@ A subsequent WRITE to the same page could be logged 0xb5 0x7f 0x23
 0x34 0x56 0x78, meaning "same page, type code 3 (WRITE), 5 bytes to
 follow", "byte offset 0x7f"+0x60+2, bytes 0x23,0x34,0x56,0x78.
 
-The end of the mini-transaction would be indicated by a NUL byte.
-*/
+The end of the mini-transaction would be indicated by the end byte
+0x00 or 0x01; @see log_sys.get_sequence_bit().
+If log_sys.is_encrypted(), that is followed by 8 bytes of nonce
+(part of initialization vector). That will be followed by 4 bytes
+of CRC-32C of the entire mini-tranasction, excluding the end byte. */
 
 /** Redo log record types. These bit patterns (3 bits) will be written
 to the redo log file, so the existing codes or their interpretation on
@@ -209,7 +205,7 @@ enum mrec_type_t
   /** Zero-initialize a page. The current byte offset (for subsequent
   records) will be reset to FIL_PAGE_TYPE. */
   INIT_PAGE= 0x10,
-  /** Insert a record into a page. FIXME: implement this! */
+  /** Extended record; @see mrec_ext_t */
   EXTENDED= 0x20,
   /** Write a string of bytes. Followed by the byte offset (unsigned,
   relative to the current byte offset, encoded in 1 to 3 bytes) and
@@ -304,18 +300,16 @@ enum mfile_type_t
   FILE_RENAME = 0xa0,
   /** Modify a file. Followed by tablespace ID and the file name. */
   FILE_MODIFY = 0xb0,
-#if 1 /* MDEV-14425 FIXME: Remove this! */
-  /** End-of-checkpoint marker. Followed by 2 dummy bytes of page identifier,
-  8 bytes of LSN, and padded with a NUL; @see SIZE_OF_FILE_CHECKPOINT. */
+  /** End-of-checkpoint marker, at the end of a mini-transaction.
+  Followed by 2 NUL bytes of page identifier and 8 bytes of LSN;
+  @see SIZE_OF_FILE_CHECKPOINT.
+  When all bytes are NUL, this is a dummy padding record. */
   FILE_CHECKPOINT = 0xf0
-#endif
 };
 
-#if 1 /* MDEV-14425 FIXME: Remove this! */
 /** Size of a FILE_CHECKPOINT record, including the trailing byte to
-terminate the mini-transaction. */
-constexpr byte SIZE_OF_FILE_CHECKPOINT= 3/*type,page_id*/ + 8/*LSN*/ + 1;
-#endif
+terminate the mini-transaction and the CRC-32C. */
+constexpr byte SIZE_OF_FILE_CHECKPOINT= 3/*type,page_id*/ + 8/*LSN*/ + 1 + 4;
 
 #ifndef UNIV_INNOCHECKSUM
 /** Types for the mlock objects to store in the mtr_t::m_memo */
@@ -339,9 +333,9 @@ enum mtr_memo_type_t {
 
 	MTR_MEMO_SX_LOCK = RW_SX_LATCH << 5,
 
-	/** acquire X-latch on fil_space_t::latch */
-	MTR_MEMO_SPACE_X_LOCK = MTR_MEMO_SX_LOCK << 1
+	/** wr_lock() on fil_space_t::latch */
+	MTR_MEMO_SPACE_X_LOCK = MTR_MEMO_SX_LOCK << 1,
+	/** rd_lock() on fil_space_t::latch */
+	MTR_MEMO_SPACE_S_LOCK = MTR_MEMO_SX_LOCK << 2
 };
-#endif /* !UNIV_CHECKSUM */
-
-#endif /* mtr0types_h */
+#endif /* !UNIV_INNOCHECKSUM */
