@@ -24,18 +24,15 @@ Mini-transaction buffer
 Created 11/26/1995 Heikki Tuuri
 *******************************************************/
 
-#include "mtr0mtr.h"
-
+#include "mtr0log.h"
 #include "buf0buf.h"
 #include "buf0flu.h"
-#include "fsp0sysspace.h"
 #include "page0types.h"
-#include "mtr0log.h"
-#include "log0recv.h"
-#include "my_cpu.h"
+#include "log0crypt.h"
 #ifdef BTR_CUR_HASH_ADAPT
 # include "btr0sea.h"
 #endif
+#include "log.h"
 
 /** Iterate over a memo block in reverse. */
 template <typename Functor>
@@ -299,50 +296,6 @@ struct ReleaseAll {
   }
 };
 
-/** Stops iteration is savepoint is reached */
-template <typename Functor> struct TillSavepoint
-{
-
-  /** Constructor
-  @param[in] functor functor which is called if savepoint is not reached
-  @param[in] savepoint savepoint value to rollback
-  @param[in] used current position in slots container */
-  TillSavepoint(const Functor &functor, ulint savepoint, ulint used)
-      : functor(functor),
-        m_slots_count((used - savepoint) / sizeof(mtr_memo_slot_t))
-  {
-    ut_ad(savepoint);
-    ut_ad(used >= savepoint);
-  }
-
-  /** @return true if savepoint is not reached, false otherwise */
-  bool operator()(mtr_memo_slot_t *slot)
-  {
-#ifdef UNIV_DEBUG
-    /** This check is added because the code is invoked only from
-    row_search_mvcc() to release latches acquired during clustered index search
-    for secondary index record. To make it more universal we could add one more
-    member in this functor for debug build to pass only certain slot types,
-    but this is currently not necessary. */
-    switch (slot->type)
-    {
-    case MTR_MEMO_S_LOCK:
-    case MTR_MEMO_PAGE_S_FIX:
-      break;
-    default:
-      ut_a(false);
-    }
-#endif
-    return m_slots_count-- && functor(slot);
-  }
-
-private:
-  /** functor to invoke */
-  const Functor &functor;
-  /** slots count left till savepoint */
-  ulint m_slots_count;
-};
-
 #ifdef UNIV_DEBUG
 /** Check that all slots have been handled. */
 struct DebugCheck {
@@ -359,10 +312,22 @@ struct DebugCheck {
 struct ReleaseBlocks
 {
   const lsn_t start, end;
-  ReleaseBlocks(lsn_t start, lsn_t end) : start(start), end(end) {}
+#ifdef UNIV_DEBUG
+  const mtr_buf_t &memo;
+
+  ReleaseBlocks(lsn_t start, lsn_t end, const mtr_buf_t &memo) :
+    start(start), end(end), memo(memo)
+#else /* UNIV_DEBUG */
+  ReleaseBlocks(lsn_t start, lsn_t end, const mtr_buf_t&) :
+    start(start), end(end)
+#endif /* UNIV_DEBUG */
+  {
+    ut_ad(start);
+    ut_ad(end);
+  }
 
   /** @return true always */
-  bool operator()(mtr_memo_slot_t *slot) const
+  bool operator()(mtr_memo_slot_t* slot) const
   {
     if (!slot->object)
       return true;
@@ -375,8 +340,8 @@ struct ReleaseBlocks
       return true;
     }
 
-    buf_block_t *block= static_cast<buf_block_t*>(slot->object);
-    buf_flush_note_modification(block, start, end);
+    buf_flush_note_modification(static_cast<buf_block_t*>(slot->object),
+                                start, end);
     return true;
   }
 };
@@ -436,7 +401,7 @@ void mtr_t::commit()
 
     std::pair<lsn_t,page_flush_ahead> lsns;
 
-    if (UNIV_LIKELY(is_logged()))
+    if (UNIV_LIKELY(m_log_mode == MTR_LOG_ALL))
     {
       lsns= do_write();
 
@@ -480,8 +445,9 @@ void mtr_t::commit()
     else
       ut_ad(!m_freed_space);
 
-    m_memo.for_each_block_in_reverse
-      (CIterate<const ReleaseBlocks>(ReleaseBlocks(lsns.first, m_commit_lsn)));
+    m_memo.for_each_block_in_reverse(CIterate<const ReleaseBlocks>
+                                     (ReleaseBlocks(lsns.first, m_commit_lsn,
+                                                    m_memo)));
     if (m_made_dirty)
       mysql_mutex_unlock(&log_sys.flush_order_mutex);
 
@@ -489,9 +455,6 @@ void mtr_t::commit()
 
     if (UNIV_UNLIKELY(lsns.second != PAGE_FLUSH_NO))
       buf_flush_ahead(m_commit_lsn, lsns.second == PAGE_FLUSH_SYNC);
-
-    if (m_made_dirty)
-      srv_stats.log_write_requests.inc();
   }
   else
     m_memo.for_each_block_in_reverse(CIterate<ReleaseAll>());
@@ -499,21 +462,6 @@ void mtr_t::commit()
   release_resources();
 }
 
-/** Release latches till savepoint. To simplify the code only
-MTR_MEMO_S_LOCK and MTR_MEMO_PAGE_S_FIX slot types are allowed to be
-released, otherwise it would be neccesary to add one more argument in the
-function to point out what slot types are allowed for rollback, and this
-would be overengineering as corrently the function is used only in one place
-in the code.
-@param savepoint   savepoint, can be obtained with get_savepoint */
-void mtr_t::rollback_to_savepoint(ulint savepoint)
-{
-  Iterate<TillSavepoint<ReleaseLatches>> iteration(
-      TillSavepoint<ReleaseLatches>(ReleaseLatches(), savepoint,
-                                    get_savepoint()));
-  m_memo.for_each_block_in_reverse(iteration);
-}
-
 /** Shrink a tablespace. */
 struct Shrink
 {
@@ -577,7 +525,6 @@ void mtr_t::commit_shrink(fil_space_t &space)
   log_write_and_flush_prepare();
 
   const lsn_t start_lsn= do_write().first;
-  ut_d(m_log.erase());
 
   mysql_mutex_lock(&log_sys.flush_order_mutex);
   /* Durably write the reduced FSP_SIZE before truncating the data file. */
@@ -610,7 +557,8 @@ void mtr_t::commit_shrink(fil_space_t &space)
   m_memo.for_each_block_in_reverse(CIterate<Shrink>{space});
 
   m_memo.for_each_block_in_reverse(CIterate<const ReleaseBlocks>
-                                   (ReleaseBlocks(start_lsn, m_commit_lsn)));
+                                   (ReleaseBlocks(start_lsn, m_commit_lsn,
+                                                  m_memo)));
   mysql_mutex_unlock(&log_sys.flush_order_mutex);
 
   mysql_mutex_lock(&fil_system.mutex);
@@ -621,135 +569,62 @@ void mtr_t::commit_shrink(fil_space_t &space)
   mysql_mutex_unlock(&fil_system.mutex);
 
   m_memo.for_each_block_in_reverse(CIterate<ReleaseLatches>());
-  srv_stats.log_write_requests.inc();
 
   release_resources();
 }
 
-/** Commit a mini-transaction that is deleting or renaming a file.
-@param space   tablespace that is being renamed or deleted
-@param name    new file name (nullptr=the file will be deleted)
-@return whether the operation succeeded */
-bool mtr_t::commit_file(fil_space_t &space, const char *name)
-{
-  ut_ad(is_active());
-  ut_ad(!is_inside_ibuf());
-  ut_ad(!high_level_read_only);
-  ut_ad(m_modifications);
-  ut_ad(!m_made_dirty);
-  ut_ad(!recv_recovery_is_on());
-  ut_ad(m_log_mode == MTR_LOG_ALL);
-  ut_ad(UT_LIST_GET_LEN(space.chain) == 1);
-
-  log_write_and_flush_prepare();
-
-  do_write();
-
-  mysql_mutex_assert_owner(&log_sys.mutex);
-
-  if (!name && space.max_lsn)
-  {
-    ut_d(space.max_lsn= 0);
-    fil_system.named_spaces.remove(space);
-  }
-
-  /* Block log_checkpoint(). */
-  mysql_mutex_lock(&buf_pool.flush_list_mutex);
-
-  /* Durably write the log for the file system operation. */
-  log_write_and_flush();
-
-  char *old_name= space.chain.start->name;
-  bool success;
-
-  if (name)
-  {
-    success= os_file_rename(innodb_data_file_key, old_name, name);
-
-    if (success)
-    {
-      mysql_mutex_lock(&fil_system.mutex);
-      space.chain.start->name= mem_strdup(name);
-      mysql_mutex_unlock(&fil_system.mutex);
-      ut_free(old_name);
-    }
-  }
-  else
-  {
-    /* Remove any additional files. */
-    if (char *cfg_name= fil_make_filepath(old_name,
-					  fil_space_t::name_type{}, CFG,
-                                          false))
-    {
-      os_file_delete_if_exists(innodb_data_file_key, cfg_name, nullptr);
-      ut_free(cfg_name);
-    }
-
-    if (FSP_FLAGS_HAS_DATA_DIR(space.flags))
-      RemoteDatafile::delete_link_file(space.name());
-
-    /* Remove the directory entry. The file will actually be deleted
-    when our caller closes the handle. */
-    os_file_delete(innodb_data_file_key, old_name);
-
-    mysql_mutex_lock(&fil_system.mutex);
-    /* Sanity checks after reacquiring fil_system.mutex */
-    ut_ad(&space == fil_space_get_by_id(space.id));
-    ut_ad(!space.referenced());
-    ut_ad(space.is_stopping());
-
-    fil_system.detach(&space, true);
-    mysql_mutex_unlock(&fil_system.mutex);
-
-    success= true;
-  }
-
-  mysql_mutex_unlock(&buf_pool.flush_list_mutex);
-  ut_d(m_log.erase());
-  release_resources();
-
-  srv_stats.log_write_requests.inc();
-  return success;
-}
-
 /** Commit a mini-transaction that did not modify any pages,
 but generated some redo log on a higher level, such as
 FILE_MODIFY records and an optional FILE_CHECKPOINT marker.
 The caller must hold log_sys.mutex.
 This is to be used at log_checkpoint().
-@param[in]	checkpoint_lsn		log checkpoint LSN, or 0 */
-void mtr_t::commit_files(lsn_t checkpoint_lsn)
+@param checkpoint_lsn   the log sequence number of a checkpoint, or 0
+@return current LSN */
+lsn_t mtr_t::commit_files(lsn_t checkpoint_lsn)
 {
-	mysql_mutex_assert_owner(&log_sys.mutex);
-	ut_ad(is_active());
-	ut_ad(!is_inside_ibuf());
-	ut_ad(m_log_mode == MTR_LOG_ALL);
-	ut_ad(!m_made_dirty);
-	ut_ad(m_memo.size() == 0);
-	ut_ad(!srv_read_only_mode);
-	ut_ad(!m_freed_space);
-	ut_ad(!m_freed_pages);
+  mysql_mutex_assert_owner(&log_sys.mutex);
+  ut_ad(is_active());
+  ut_ad(!is_inside_ibuf());
+  ut_ad(m_log_mode == MTR_LOG_ALL);
+  ut_ad(!m_made_dirty);
+  ut_ad(m_memo.size() == 0);
+  ut_ad(!srv_read_only_mode);
+  ut_ad(!m_freed_space);
+  ut_ad(!m_freed_pages);
+  ut_ad(!m_user_space);
 
-	if (checkpoint_lsn) {
-		byte*	ptr = m_log.push<byte*>(SIZE_OF_FILE_CHECKPOINT);
-		compile_time_assert(SIZE_OF_FILE_CHECKPOINT == 3 + 8 + 1);
-		*ptr = FILE_CHECKPOINT | (SIZE_OF_FILE_CHECKPOINT - 2);
-		::memset(ptr + 1, 0, 2);
-		mach_write_to_8(ptr + 3, checkpoint_lsn);
-		ptr[3 + 8] = 0;
-	} else {
-		*m_log.push<byte*>(1) = 0;
-	}
+  if (checkpoint_lsn)
+  {
+    byte *ptr= m_log.push<byte*>(3 + 8);
+    *ptr= FILE_CHECKPOINT | (2 + 8);
+    ::memset(ptr + 1, 0, 2);
+    mach_write_to_8(ptr + 3, checkpoint_lsn);
+  }
 
-	finish_write(m_log.size());
-	srv_stats.log_write_requests.inc();
-	release_resources();
+  size_t size= m_log.size() + 5;
 
-	if (checkpoint_lsn) {
-		DBUG_PRINT("ib_log",
-			   ("FILE_CHECKPOINT(" LSN_PF ") written at " LSN_PF,
-			    checkpoint_lsn, log_sys.get_lsn()));
-	}
+  if (log_sys.is_encrypted())
+  {
+    /* We will not encrypt any FILE_ records, but we will reserve
+    a nonce at the end. */
+    size+= 8;
+    m_commit_lsn= log_sys.get_lsn();
+  }
+  else
+    m_commit_lsn= 0;
+
+  m_crc= 0;
+  m_log.for_each_block([this](const mtr_buf_t::block_t *b)
+  { m_crc= my_crc32c(m_crc, b->begin(), b->used()); return true; });
+  finish_write(size);
+  release_resources();
+
+  if (checkpoint_lsn)
+    DBUG_PRINT("ib_log",
+               ("FILE_CHECKPOINT(" LSN_PF ") written at " LSN_PF,
+                checkpoint_lsn, m_commit_lsn));
+
+  return m_commit_lsn;
 }
 
 #ifdef UNIV_DEBUG
@@ -760,9 +635,19 @@ void mtr_t::commit_files(lsn_t checkpoint_lsn)
 bool
 mtr_t::is_named_space(uint32_t space) const
 {
-  ut_ad(!m_user_space || m_user_space->id != TRX_SYS_SPACE);
-  return !is_logged() || m_user_space_id == space ||
-    is_predefined_tablespace(space);
+	ut_ad(!m_user_space || m_user_space->id != TRX_SYS_SPACE);
+
+	switch (m_log_mode) {
+	case MTR_LOG_NONE:
+	case MTR_LOG_NO_REDO:
+		return(true);
+	case MTR_LOG_ALL:
+		return(m_user_space_id == space
+		       || is_predefined_tablespace(space));
+	}
+
+	ut_error;
+	return(false);
 }
 /** Check if a tablespace is associated with the mini-transaction
 (needed for generating a FILE_MODIFY record)
@@ -772,8 +657,16 @@ bool mtr_t::is_named_space(const fil_space_t* space) const
 {
   ut_ad(!m_user_space || m_user_space->id != TRX_SYS_SPACE);
 
-  return !is_logged() || m_user_space == space ||
-    is_predefined_tablespace(space->id);
+  switch (m_log_mode) {
+  case MTR_LOG_NONE:
+  case MTR_LOG_NO_REDO:
+    return true;
+  case MTR_LOG_ALL:
+    return m_user_space == space || is_predefined_tablespace(space->id);
+  }
+
+  ut_error;
+  return false;
 }
 #endif /* UNIV_DEBUG */
 
@@ -865,32 +758,32 @@ static time_t log_margin_warn_time;
 static bool log_close_warned;
 static time_t log_close_warn_time;
 
-/** Check margin not to overwrite transaction log from the last checkpoint.
-If would estimate the log write to exceed the log_capacity,
-waits for the checkpoint is done enough.
-@param len   length of the data to be written */
-static void log_margin_checkpoint_age(ulint len)
+/** Display a warning that the log tail is overwriting the head,
+making the server crash-unsafe. */
+ATTRIBUTE_COLD static void log_overwrite_warning(lsn_t age, lsn_t capacity)
 {
-  const ulint framing_size= log_sys.framing_size();
-  /* actual length stored per block */
-  const ulint len_per_blk= OS_FILE_LOG_BLOCK_SIZE - framing_size;
+  time_t t= time(nullptr);
+  if (!log_close_warned || difftime(t, log_close_warn_time) > 15)
+  {
+    log_close_warned= true;
+    log_close_warn_time= t;
 
-  /* actual data length in last block already written */
-  ulint extra_len= log_sys.buf_free % OS_FILE_LOG_BLOCK_SIZE;
+    sql_print_error("InnoDB: The age of the last checkpoint is " LSN_PF
+                    ", which exceeds the log capacity " LSN_PF ".",
+                    age, capacity);
+  }
+}
 
-  ut_ad(extra_len >= LOG_BLOCK_HDR_SIZE);
-  extra_len-= LOG_BLOCK_HDR_SIZE;
+/** Reserve space in the log buffer for appending data.
+@param size   upper limit of the length of the data to append(), in bytes
+@return the current LSN */
+inline lsn_t log_t::append_prepare(size_t size) noexcept
+{
+  mysql_mutex_assert_owner(&mutex);
 
-  /* total extra length for block header and trailer */
-  extra_len= ((len + extra_len) / len_per_blk) * framing_size;
+  lsn_t lsn= get_lsn();
 
-  const ulint margin= len + extra_len;
-
-  mysql_mutex_assert_owner(&log_sys.mutex);
-
-  const lsn_t lsn= log_sys.get_lsn();
-
-  if (UNIV_UNLIKELY(margin > log_sys.log_capacity))
+  if (UNIV_UNLIKELY(size > log_capacity))
   {
     time_t t= time(nullptr);
 
@@ -900,143 +793,63 @@ static void log_margin_checkpoint_age(ulint len)
       log_margin_warned= true;
       log_margin_warn_time= t;
 
-      ib::error() << "innodb_log_file_size is too small "
-                     "for mini-transaction size " << len;
+      sql_print_error("InnoDB: innodb_log_file_size is too small "
+                      "for mini-transaction size %zu", size);
     }
+    goto throttle;
   }
-  else if (UNIV_LIKELY(lsn + margin <= log_sys.last_checkpoint_lsn +
-                       log_sys.log_capacity))
-    return;
+  else if (UNIV_UNLIKELY(lsn + size > last_checkpoint_lsn + log_capacity))
+  throttle:
+    set_check_flush_or_checkpoint();
 
-  log_sys.set_check_flush_or_checkpoint();
-}
-
-
-/** Open the log for log_write_low(). The log must be closed with log_close().
-@param len length of the data to be written
-@return start lsn of the log record */
-static lsn_t log_reserve_and_open(size_t len)
-{
-  for (ut_d(ulint count= 0);;)
+  if (is_pmem())
   {
-    mysql_mutex_assert_owner(&log_sys.mutex);
+    for (ut_d(int count= 50); capacity() - size <
+           size_t(lsn - flushed_to_disk_lsn.load(std::memory_order_relaxed)); )
+    {
+      waits++;
+      mysql_mutex_unlock(&mutex);
+      DEBUG_SYNC_C("log_buf_size_exceeded");
+      log_write_up_to(lsn, true);
+      ut_ad(count--);
+      mysql_mutex_lock(&mutex);
+      lsn= get_lsn();
+    }
+    return lsn;
+  }
 
-    /* Calculate an upper limit for the space the string may take in
-    the log buffer */
+  /* Calculate the amount of free space needed. */
+  size= (4 * 4096) - size + log_sys.buf_size;
 
-    size_t len_upper_limit= (4 * OS_FILE_LOG_BLOCK_SIZE) +
-      srv_log_write_ahead_size + (5 * len) / 4;
-
-    if (log_sys.buf_free + len_upper_limit <= srv_log_buffer_size)
-      break;
-
-    mysql_mutex_unlock(&log_sys.mutex);
+  for (ut_d(int count= 50); UNIV_UNLIKELY(buf_free > size); )
+  {
+    waits++;
+    mysql_mutex_unlock(&mutex);
     DEBUG_SYNC_C("log_buf_size_exceeded");
-
-    /* Not enough free space, do a write of the log buffer */
-    log_write_up_to(log_sys.get_lsn(), false);
-
-    srv_stats.log_waits.inc();
-
-    ut_ad(++count < 50);
-
-    mysql_mutex_lock(&log_sys.mutex);
+    log_write_up_to(lsn, false);
+    ut_ad(count--);
+    mysql_mutex_lock(&mutex);
+    lsn= get_lsn();
   }
 
-  return log_sys.get_lsn();
+  return lsn;
 }
 
-/** Append data to the log buffer. */
-static void log_write_low(const void *str, size_t size)
+/** Finish appending data to the log.
+@param lsn  the end LSN of the log record
+@return whether buf_flush_ahead() will have to be invoked */
+static mtr_t::page_flush_ahead log_close(lsn_t lsn) noexcept
 {
   mysql_mutex_assert_owner(&log_sys.mutex);
-  const ulint trailer_offset= log_sys.trailer_offset();
-
-  do
-  {
-    /* Calculate a part length */
-    size_t len= size;
-    size_t data_len= (log_sys.buf_free % OS_FILE_LOG_BLOCK_SIZE) + size;
-
-    if (data_len > trailer_offset)
-    {
-      data_len= trailer_offset;
-      len= trailer_offset - log_sys.buf_free % OS_FILE_LOG_BLOCK_SIZE;
-    }
-
-    memcpy(log_sys.buf + log_sys.buf_free, str, len);
-
-    size-= len;
-    str= static_cast<const char*>(str) + len;
-
-    byte *log_block= static_cast<byte*>(ut_align_down(log_sys.buf +
-                                                      log_sys.buf_free,
-                                                      OS_FILE_LOG_BLOCK_SIZE));
-
-    log_block_set_data_len(log_block, data_len);
-    lsn_t lsn= log_sys.get_lsn();
-
-    if (data_len == trailer_offset)
-    {
-      /* This block became full */
-      log_block_set_data_len(log_block, OS_FILE_LOG_BLOCK_SIZE);
-      log_block_set_checkpoint_no(log_block, log_sys.next_checkpoint_no);
-      len+= log_sys.framing_size();
-      lsn+= len;
-      /* Initialize the next block header */
-      log_block_init(log_block + OS_FILE_LOG_BLOCK_SIZE, lsn);
-    }
-    else
-      lsn+= len;
-
-    log_sys.set_lsn(lsn);
-    log_sys.buf_free+= len;
-
-    ut_ad(log_sys.buf_free <= size_t{srv_log_buffer_size});
-  }
-  while (size);
-}
-
-/** Close the log at mini-transaction commit.
-@return whether buffer pool flushing is needed */
-static mtr_t::page_flush_ahead log_close(lsn_t lsn)
-{
-  mysql_mutex_assert_owner(&log_sys.mutex);
-  ut_ad(lsn == log_sys.get_lsn());
-
-  byte *log_block= static_cast<byte*>(ut_align_down(log_sys.buf +
-                                                    log_sys.buf_free,
-                                                    OS_FILE_LOG_BLOCK_SIZE));
-
-  if (!log_block_get_first_rec_group(log_block))
-  {
-    /* We initialized a new log block which was not written
-    full by the current mtr: the next mtr log record group
-    will start within this block at the offset data_len */
-    log_block_set_first_rec_group(log_block,
-                                  log_block_get_data_len(log_block));
-  }
-
-  if (log_sys.buf_free > log_sys.max_buf_free)
-    log_sys.set_check_flush_or_checkpoint();
+  log_sys.write_to_buf++;
+  log_sys.set_lsn(lsn);
 
   const lsn_t checkpoint_age= lsn - log_sys.last_checkpoint_lsn;
 
   if (UNIV_UNLIKELY(checkpoint_age >= log_sys.log_capacity) &&
       /* silence message on create_log_file() after the log had been deleted */
       checkpoint_age != lsn)
-  {
-    time_t t= time(nullptr);
-    if (!log_close_warned || difftime(t, log_close_warn_time) > 15)
-    {
-      log_close_warned= true;
-      log_close_warn_time= t;
-
-      ib::error() << "The age of the last checkpoint is " << checkpoint_age
-                  << ", which exceeds the log capacity "
-                  << log_sys.log_capacity << ".";
-    }
-  }
+    log_overwrite_warning(checkpoint_age, log_sys.log_capacity);
   else if (UNIV_LIKELY(checkpoint_age <= log_sys.max_modified_age_async))
     return mtr_t::PAGE_FLUSH_NO;
   else if (UNIV_LIKELY(checkpoint_age <= log_sys.max_checkpoint_age))
@@ -1046,163 +859,125 @@ static mtr_t::page_flush_ahead log_close(lsn_t lsn)
   return mtr_t::PAGE_FLUSH_SYNC;
 }
 
-inline void mtr_t::page_checksum(const buf_page_t &bpage)
-{
-  const byte *page= bpage.frame;
-  size_t size= srv_page_size;
-
-  if (UNIV_LIKELY_NULL(bpage.zip.data))
-  {
-    size= (UNIV_ZIP_SIZE_MIN >> 1) << bpage.zip.ssize;
-    switch (fil_page_get_type(bpage.zip.data)) {
-    case FIL_PAGE_TYPE_ALLOCATED:
-    case FIL_PAGE_INODE:
-    case FIL_PAGE_IBUF_BITMAP:
-    case FIL_PAGE_TYPE_FSP_HDR:
-    case FIL_PAGE_TYPE_XDES:
-      /* These are essentially uncompressed pages. */
-      break;
-    default:
-      page= bpage.zip.data;
-    }
-  }
-
-  /* We have to exclude from the checksum the normal
-  page checksum that is written by buf_flush_init_for_writing()
-  and FIL_PAGE_LSN which would be updated once we have actually
-  allocated the LSN.
-
-  Unfortunately, we cannot access fil_space_t easily here. In order to
-  be compatible with encrypted tablespaces in the pre-full_crc32
-  format we will unconditionally exclude the 8 bytes at
-  FIL_PAGE_FILE_FLUSH_LSN_OR_KEY_VERSION
-  a.k.a. FIL_RTREE_SPLIT_SEQ_NUM. */
-  const uint32_t checksum=
-    my_crc32c(my_crc32c(my_crc32c(0, page + FIL_PAGE_OFFSET,
-                                  FIL_PAGE_LSN - FIL_PAGE_OFFSET),
-                        page + FIL_PAGE_TYPE, 2),
-              page + FIL_PAGE_SPACE_ID, size - (FIL_PAGE_SPACE_ID + 8));
-
-  byte *l= log_write<OPTION>(bpage.id(), nullptr, 5, true, 0);
-  *l++= OPT_PAGE_CHECKSUM;
-  mach_write_to_4(l, checksum);
-  m_log.close(l + 4);
-}
-
-/** Write OPT_PAGE_CHECKSUM records for modified pages */
-struct WriteOPT_PAGE_CHECKSUM
-{
-  mtr_t &mtr;
-  WriteOPT_PAGE_CHECKSUM(mtr_t &mtr) : mtr(mtr) {}
-
-  /** @return true always */
-  bool operator()(const mtr_memo_slot_t *slot) const
-  {
-    if (slot->type & MTR_MEMO_MODIFY)
-    {
-      const buf_page_t &b= static_cast<const buf_block_t*>(slot->object)->page;
-      if (!b.is_freed())
-        mtr.page_checksum(b);
-    }
-    return true;
-  }
-};
-
-/** Write the block contents to the REDO log */
-struct mtr_write_log
-{
-  /** Append a block to the redo log buffer.
-  @return whether the appending should continue */
-  bool operator()(const mtr_buf_t::block_t *block) const
-  {
-    log_write_low(block->begin(), block->used());
-    return true;
-  }
-};
-
 std::pair<lsn_t,mtr_t::page_flush_ahead> mtr_t::do_write()
 {
-	ut_ad(!recv_no_log_write);
-	ut_ad(is_logged());
+  ut_ad(!recv_no_log_write);
+  ut_ad(m_log_mode == MTR_LOG_ALL);
 
-	ulint	len	= m_log.size();
-	ut_ad(len > 0);
+  size_t len= m_log.size() + 5;
+  ut_ad(len > 5);
+  if (log_sys.is_encrypted())
+  {
+    len+= 8;
+    encrypt();
+  }
+  else
+  {
+    m_crc= 0;
+    m_commit_lsn= 0;
+    m_log.for_each_block([this](const mtr_buf_t::block_t *b)
+    { m_crc= my_crc32c(m_crc, b->begin(), b->used()); return true; });
+  }
 
-#ifndef DBUG_OFF
-	if (m_log_mode == MTR_LOG_ALL) {
-		do {
-			DBUG_EXECUTE_IF("skip_page_checksum", continue;);
-			m_memo.for_each_block(CIterate<WriteOPT_PAGE_CHECKSUM>
-					      (*this));
-			len = m_log.size();
-		} while (0);
-	}
-#endif
+  mysql_mutex_lock(&log_sys.mutex);
 
-	if (len > srv_log_buffer_size / 2) {
-		log_buffer_extend(ulong((len + 1) * 2));
-	}
+  if (m_user_space && !is_predefined_tablespace(m_user_space->id) &&
+      !m_user_space->max_lsn)
+    name_write();
 
-	fil_space_t*	space = m_user_space;
-
-	if (space != NULL && is_predefined_tablespace(space->id)) {
-		/* Omit FILE_MODIFY for predefined tablespaces. */
-		space = NULL;
-	}
-
-	mysql_mutex_lock(&log_sys.mutex);
-
-	if (fil_names_write_if_was_clean(space)) {
-		len = m_log.size();
-	} else {
-		/* This was not the first time of dirtying a
-		tablespace since the latest checkpoint. */
-		ut_ad(len == m_log.size());
-	}
-
-	*m_log.push<byte*>(1) = 0;
-	len++;
-
-	/* check and attempt a checkpoint if exceeding capacity */
-	log_margin_checkpoint_age(len);
-
-	return finish_write(len);
+  return finish_write(len);
 }
 
-/** Append the redo log records to the redo log buffer.
-@param len   number of bytes to write
+/** Write the mini-transaction log to the redo log buffer.
 @return {start_lsn,flush_ahead} */
-inline std::pair<lsn_t,mtr_t::page_flush_ahead> mtr_t::finish_write(ulint len)
+std::pair<lsn_t,mtr_t::page_flush_ahead> mtr_t::finish_write(size_t len)
 {
-	ut_ad(is_logged());
-	mysql_mutex_assert_owner(&log_sys.mutex);
-	ut_ad(m_log.size() == len);
-	ut_ad(len > 0);
+  ut_ad(!recv_no_log_write);
+  ut_ad(m_log_mode == MTR_LOG_ALL);
 
-	lsn_t start_lsn;
+  const lsn_t start_lsn= log_sys.append_prepare(len);
+  const size_t size{m_commit_lsn ? 5U + 8U : 5U};
 
-	if (m_log.is_small()) {
-		const mtr_buf_t::block_t* front = m_log.front();
-		ut_ad(len <= front->used());
+  if (!log_sys.is_pmem())
+  {
+    m_log.for_each_block([](const mtr_buf_t::block_t *b)
+    { log_sys.append(b->begin(), b->used()); return true; });
 
-		m_commit_lsn = log_reserve_and_write_fast(front->begin(), len,
-							  &start_lsn);
+    if (log_sys.buf_free >= log_sys.max_buf_free)
+      log_sys.set_check_flush_or_checkpoint();
 
-		if (!m_commit_lsn) {
-			goto piecewise;
-		}
-	} else {
-piecewise:
-		/* Open the database log for log_write_low */
-		start_lsn = log_reserve_and_open(len);
-		mtr_write_log write_log;
-		m_log.for_each_block(write_log);
-		m_commit_lsn = log_sys.get_lsn();
-	}
-	page_flush_ahead flush= log_close(m_commit_lsn);
-	DBUG_EXECUTE_IF("ib_log_flush_ahead", flush = PAGE_FLUSH_SYNC;);
+#ifdef HAVE_PMEM
+  write_trailer:
+#endif
+    log_sys.buf[log_sys.buf_free]=
+      log_sys.get_sequence_bit(start_lsn + len - size);
+    if (m_commit_lsn)
+    {
+      byte *nonce= log_sys.buf + log_sys.buf_free + 1;
+      mach_write_to_8(nonce, m_commit_lsn);
+      m_crc= my_crc32c(m_crc, nonce, 8);
+      mach_write_to_4(&log_sys.buf[log_sys.buf_free + 9], m_crc);
+      log_sys.buf_free+= 8 + 5;
+    }
+    else
+    {
+      mach_write_to_4(&log_sys.buf[log_sys.buf_free + 1], m_crc);
+      log_sys.buf_free+= 5;
+    }
+  }
+#ifdef HAVE_PMEM
+  else if (UNIV_LIKELY(log_sys.buf_free + len < log_sys.file_size))
+  {
+    m_log.for_each_block([](const mtr_buf_t::block_t *b)
+    { log_sys.append(b->begin(), b->used()); return true; });
+    goto write_trailer;
+  }
+  else
+  {
+    m_log.for_each_block([](const mtr_buf_t::block_t *b)
+    {
+      size_t size{b->used()};
+      const size_t size_left{log_sys.file_size - log_sys.buf_free};
+      const byte *src= b->begin();
+      if (size <= size_left)
+      {
+        ::memcpy(log_sys.buf + log_sys.buf_free, src, size);
+        log_sys.buf_free+= size;
+      }
+      else
+      {
+        size-= size_left;
+        ::memcpy(log_sys.buf + log_sys.buf_free, src, size_left);
+        ::memcpy(log_sys.buf + log_sys.START_OFFSET, src + size_left, size);
+        log_sys.buf_free= log_sys.START_OFFSET + size;
+      }
+      return true;
+    });
+    const size_t size_left{log_sys.file_size - log_sys.buf_free};
+    if (size_left > size)
+      goto write_trailer;
 
-	return std::make_pair(start_lsn, flush);
+    byte tail[5 + 8];
+    tail[0]= log_sys.get_sequence_bit(start_lsn + len - size);
+
+    if (m_commit_lsn)
+    {
+      mach_write_to_8(tail + 1, m_commit_lsn);
+      m_crc= my_crc32c(m_crc, tail + 1, 8);
+      mach_write_to_4(tail + 9, m_crc);
+    }
+    else
+      mach_write_to_4(tail + 1, m_crc);
+
+    ::memcpy(log_sys.buf + log_sys.buf_free, tail, size_left);
+    ::memcpy(log_sys.buf + log_sys.START_OFFSET, tail + size_left,
+             size - size_left);
+    log_sys.buf_free= log_sys.START_OFFSET + (size - size_left);
+  }
+#endif
+
+  m_commit_lsn= start_lsn + len;
+  return {start_lsn, log_close(m_commit_lsn)};
 }
 
 /** Find out whether a block was not X-latched by the mini-transaction */
@@ -1215,7 +990,7 @@ struct FindBlockX
   /** @return whether the block was not found x-latched */
   bool operator()(const mtr_memo_slot_t *slot) const
   {
-    return slot->object != &block || !(slot->type & MTR_MEMO_PAGE_X_FIX);
+    return slot->object != &block || slot->type != MTR_MEMO_PAGE_X_FIX;
   }
 };
 
@@ -1265,17 +1040,18 @@ bool mtr_t::memo_contains(const fil_space_t& space, bool shared)
 }
 
 #ifdef BTR_CUR_HASH_ADAPT
-/** If a stale adaptive hash index exists on the block, drop it. */
+/** If a stale adaptive hash index exists on the block, drop it.
+Multiple executions of btr_search_drop_page_hash_index() on the
+same block must be prevented by exclusive page latch. */
 ATTRIBUTE_COLD
-void mtr_t::defer_drop_ahi(buf_block_t *block, mtr_memo_type_t fix_type)
+static void mtr_defer_drop_ahi(buf_block_t *block, mtr_memo_type_t fix_type)
 {
   switch (fix_type) {
-  default:
-    ut_ad(fix_type == MTR_MEMO_BUF_FIX);
-    /* We do not drop the adaptive hash index, because safely doing so
-    would require acquiring exclusive block->page.lock, which could
-    lead to hangs in some access paths. Those code paths should have
-    no business accessing the adaptive hash index anyway. */
+  case MTR_MEMO_BUF_FIX:
+    /* We do not drop the adaptive hash index, because safely doing
+    so would require acquiring block->lock, and that is not safe
+    to acquire in some RW_NO_LATCH access paths. Those code paths
+    should have no business accessing the adaptive hash index anyway. */
     break;
   case MTR_MEMO_PAGE_S_FIX:
     /* Temporarily release our S-latch. */
@@ -1295,7 +1071,8 @@ void mtr_t::defer_drop_ahi(buf_block_t *block, mtr_memo_type_t fix_type)
         btr_search_drop_page_hash_index(block);
     block->page.lock.x_u_downgrade();
     break;
-  case MTR_MEMO_PAGE_X_FIX:
+  default:
+    ut_ad(fix_type == MTR_MEMO_PAGE_X_FIX);
     btr_search_drop_page_hash_index(block);
   }
 }
@@ -1384,7 +1161,7 @@ void mtr_t::page_lock(buf_block_t *block, ulint rw_latch)
 #ifdef BTR_CUR_HASH_ADAPT
   if (dict_index_t *index= block->index)
     if (index->freed())
-      defer_drop_ahi(block, fix_type);
+      mtr_defer_drop_ahi(block, fix_type);
 #endif /* BTR_CUR_HASH_ADAPT */
 
 done:
@@ -1519,10 +1296,20 @@ mtr_t::memo_contains_page_flagged(
 	return m_memo.for_each_block_in_reverse(iteration)
 		? NULL : iteration.functor.get_block();
 }
+
+/** Print info of an mtr handle. */
+void
+mtr_t::print() const
+{
+	ib::info() << "Mini-transaction handle: memo size "
+		<< m_memo.size() << " bytes log size "
+		<< get_log()->size() << " bytes";
+}
+
 #endif /* UNIV_DEBUG */
 
 
-/** Find a potentially modified block. */
+/** Find a block, preferrably in MTR_MEMO_MODIFY state */
 struct FindModified
 {
   mtr_memo_slot_t *found= nullptr;
@@ -1534,7 +1321,8 @@ struct FindModified
     if (slot->object != &block)
       return true;
     found= slot;
-    return !(slot->type & (MTR_MEMO_PAGE_X_FIX | MTR_MEMO_PAGE_SX_FIX));
+    return !(slot->type & (MTR_MEMO_MODIFY |
+                           MTR_MEMO_PAGE_X_FIX | MTR_MEMO_PAGE_SX_FIX));
   }
 };
 
@@ -1550,73 +1338,12 @@ void mtr_t::modify(const buf_block_t &block)
   }
 
   Iterate<FindModified> iteration((FindModified(block)));
-  if (UNIV_UNLIKELY(m_memo.for_each_block(iteration)))
+  m_memo.for_each_block(iteration);
+  if (UNIV_UNLIKELY(!iteration.functor.found))
   {
     ut_ad("modifying an unlatched page" == 0);
     return;
   }
   iteration.functor.found->type= static_cast<mtr_memo_type_t>
     (iteration.functor.found->type | MTR_MEMO_MODIFY);
-  if (is_block_dirtied(&block))
-    m_made_dirty= true;
-}
-
-/** Handle an exclusively latched block that was later marked as freed. */
-struct MarkFreed
-{
-  const page_id_t id;
-  mutable buf_block_t *freed= nullptr;
-  MarkFreed(page_id_t id) : id(id) {}
-
-  bool operator()(mtr_memo_slot_t *slot) const
-  {
-    buf_block_t *block= static_cast<buf_block_t*>(slot->object);
-    if (!block);
-    else if (block == freed)
-    {
-      if (slot->type & (MTR_MEMO_PAGE_SX_FIX | MTR_MEMO_PAGE_X_FIX))
-        slot->type= MTR_MEMO_PAGE_X_FIX;
-      else
-      {
-        ut_ad(slot->type == MTR_MEMO_BUF_FIX);
-        block->page.unfix();
-        slot->object= nullptr;
-      }
-    }
-    else if (slot->type & (MTR_MEMO_PAGE_X_FIX | MTR_MEMO_PAGE_SX_FIX) &&
-             block->page.id() == id)
-    {
-      ut_ad(!block->page.is_freed());
-      ut_ad(!freed);
-      freed= block;
-      if (!(slot->type & MTR_MEMO_PAGE_X_FIX))
-      {
-        ut_d(bool upgraded=) block->page.lock.x_lock_upgraded();
-        ut_ad(upgraded);
-      }
-      slot->type= MTR_MEMO_PAGE_X_MODIFY;
-#ifdef BTR_CUR_HASH_ADAPT
-      if (block->index)
-        btr_search_drop_page_hash_index(block);
-#endif /* BTR_CUR_HASH_ADAPT */
-      block->page.set_freed(block->page.state());
-    }
-    return true;
-  }
-};
-
-/** Free a page.
-@param space   tablespace
-@param offset  offset of the page to be freed */
-void mtr_t::free(const fil_space_t &space, uint32_t offset)
-{
-  ut_ad(is_named_space(&space));
-  ut_ad(!m_freed_space || m_freed_space == &space);
-
-  if (is_logged())
-  {
-    m_memo.for_each_block_in_reverse
-      (CIterate<MarkFreed>((MarkFreed{{space.id, offset}})));
-    m_log.close(log_write<FREE_PAGE>({space.id, offset}, nullptr));
-  }
 }
